"""
Remarks on Internal Packages Imports:
    from sfft.utils.HoughDetection import Hough_Detection
    from sfft.utils.WeightedQuantile import TopFlatten_Weighted_Quantile
    from sfft.utils.pyAstroMatic.PYSEx import PY_SEx

"""

from .ReadWCS import Read_WCS
from .CombineHeader import Combine_Header
from .SymmetricMatch import Symmetric_Match
from .StampGenerator import Stamp_Generator
from .SExSkySubtract import SEx_SkySubtract
from .HoughDetection import Hough_Detection
from .PureCupyFFTKits import PureCupy_FFTKits
from .SkyLevelEstimator import SkyLevel_Estimator
from .HoughMorphClassifier import Hough_MorphClassifier
from .ConvKernelConvertion import ConvKernel_Convertion
from .DeCorrelationCalculator import DeCorrelation_Calculator
from .PatternRotationCalculator import PatternRotation_Calculator
from .NeighboringPixelCovariance import NeighboringPixel_Covariance
from .PureCupyDeCorrelationCalculator import PureCupy_DeCorrelation_Calculator
from .WeightedQuantile import Weighted_Quantile, TopFlatten_Weighted_Quantile
from .SFFTSolutionReader import Read_SFFTSolution, SVKDict_ST2SFFT, SVKDict_SFFT2ST, Realize_MatchingKernel, Realize_FluxScaling



import os
import warnings
import os.path as pa
from astropy.io import fits
from tempfile import mkdtemp
from sfft.utils.ReadWCS import Read_WCS
from sfft.utils.CombineHeader import Combine_Header
from sfft.utils.pyAstroMatic.AMConfigMaker import AMConfig_Maker
# version: Apr 22, 2024

# improved by Lauren Aldoroty (Duke Univ.)
__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class PY_SWarp:
    @staticmethod
    def PS(FITS_obj, FITS_ref, FITS_resamp=None, \
        GAIN_KEY='GAIN', SATUR_KEY='SATURATE', GAIN_DEFAULT=0.0, SATLEV_DEFAULT=50000.0, \
        OVERSAMPLING=1, RESAMPLING_TYPE='LANCZOS3', SUBTRACT_BACK='N', FILL_VALUE=None, \
        TMPDIR_ROOT=None, VERBOSE_TYPE='NORMAL', VERBOSE_LEVEL=2):
        
        """
        # Inputs & Outputs:

        -FITS_obj []                        # FITS file path of the input image to be resampled

        -FITS_ref []                        # FITS file path of the input image as resampling reference

        -FITS_resamp [None]                 # FITS file path of the output image of resampled -FITS_obj

        # SWarp parameters:

        -GAIN_KEY ['GAIN']                  # SWarp Parameter GAIN_KEYWORD
                                            # i.e., keyword of GAIN in FITS image header

        -SATUR_KEY ['SATURATE']             # SWarp Parameter SATLEV_KEYWORD
                                            # i.e., keyword of the saturation level in the FITS image header

        -GAIN_DEFAULT [0.0]                 # Gain value if no header keyword available

        -SATLEV_DEFAULT [50000.0]           # Saturation value if no header keyword available

        -OVERSAMPLING [1]                   # SWarp Parameter OVERSAMPLING
                                            # Oversampling in each dimension
                                            # (0 = automatic)
                                            # P.S. Here I changed the default value from 0 to 1
                                            # NOTE: large OVERSAMPLING may cause higher pixel correlation

        -RESAMPLING_TYPE ['LANCZOS3']       # SWarp Parameter RESAMPLING_TYPE
                                            # NEAREST,BILINEAR,LANCZOS2,LANCZOS3
                                            # LANCZOS4 (1 per axis) or FLAGS
                                            # NOTE: LANCZOS4 is relatively time-consuming

        -SUBTRACT_BACK ['N']                # SWarp Parameter SUBTRACT_BACK 
                                            # Subtraction sky background (Y/N)? (all or for each image)
                                            # P.S. Here I changed the default value from 'Y' to 'N'
                                       
        -VERBOSE_TYPE ['NORMAL']            # SWarp Parameter VERBOSE_TYPE
                                            # QUIET,LOG,NORMAL, or FULL

        # Other parameters:

        -FILL_VALUE [None]                  # How to fill the invalid (boundary) pixels in -FITS_resamp and -FITS_resamp_weight
                                            # e.g., -FILL_VALUE = 0.0, -FILL_VALUE = np.nan

        -VERBOSE_LEVEL [2]                  # The level of verbosity, can be [0, 1, 2]
                                            # 0/1/2: QUIET/NORMAL/FULL mode
                                            # NOTE: it only controls the verbosity out of SWarp.

        -TMPDIR_ROOT [None]                 # Specify root directory of temporary working directory. 

        # Returns:

            PixA_resamp                     # Pixel array of the resampled image
                                            # P.S. PixA_resamp = fits.getdata(FITS_resamp, ext=0).T
            
            MissingMask                    # Pixel array of the invalid pixels
                                            # P.S. -FILL_VALUE will fill the pixels where MissingMask is True
        
        # * Remarks on PYSWarp
        #   [1] SWarp itself can perform resampling and subsequent image combination.
        #       However, the current python wrapper PYSWarp only performs image resampling.
        #     
        #   [2] In SWarp, one need provide a target WCS-frame stored in fname.head, with name coincident with FITS_resamp.
        #       If no target WCS-frame is available, the output image will have a canonical N-E orientation.
        #       However, the current python wrapper PYSWarp must read a target WCS-frame from a given FITS_ref.
        #
        #   [3] SWarp allows users to feed a corresponding weight map as input.
        #       However, the current python wrapper PYSWarp does not support this feature.
        #
        #   [4] SWarp will automatically calculate the following keywords and set them into the output image.
        #       > EXPTIME: sum of exposure times in the part of coadd with the most overlaps.
        #       > GAIN: effective gain with flux and weighting applied.
        #         Note that any change in pixel scale only redistribute the flux with total flux conservative,
        #         therefore, the image resampling does not alter the GAIN value.
        #       > SATURATE: Minimum of all input saturation levels with flux scaling applied.
        #         Note that a change in pixel scale can alter the saturation level.
        #       > MJD-OBS: MJD of earliest start of exposures 
        #       
        #       For PYSWarp which only performs image resampling:
        #       > We can safely preserve EXPTIME, GAIN, MJD-OBS, and use SWarp output SATURATE value.
        #       > The resulting header of the PYSWarp output is constructed by combining
        #         FITS_obj header with WCS keywords removed & FITS_ref header only with WCS keywords
        #         In additional, we will update SATURATE value as described above, and 
        #         record the SWarp parameters used in the function.
        #
        """

        # * check FITS_obj header
        objname = pa.basename(FITS_obj)
        phr_obj = fits.getheader(FITS_obj, ext=0)

        # a mild warning
        if VERBOSE_LEVEL in [1, 2]:
            if 'EXPTIME' not in phr_obj:
                _warn_message = 'SWarp cannot find keyword EXPTIME in FITS header of [%s]!' %objname
                warnings.warn('MeLOn WARNING: %s' %_warn_message)
        
        # a mild warning
        if VERBOSE_LEVEL in [1, 2]:
            if 'MJD-OBS' not in phr_obj:
                _warn_message = 'SWarp cannot find keyword MJD-OBS in FITS header of [%s]!' %objname
                warnings.warn('MeLOn WARNING: %s' %_warn_message)

        # * make directory as a workplace
        TDIR = mkdtemp(suffix=None, prefix='PYSWarp_', dir=TMPDIR_ROOT)

        # * create SWarp configuration file in TDIR
        ConfigDict = {}
        ConfigDict['GAIN_KEYWORD'] = '%s' %GAIN_KEY
        ConfigDict['SATLEV_KEYWORD'] = '%s' %SATUR_KEY

        ConfigDict['GAIN_DEFAULT'] = '%s' %GAIN_DEFAULT
        ConfigDict['SATLEV_DEFAULT'] = '%s' %SATLEV_DEFAULT

        ConfigDict['OVERSAMPLING'] = '%d' %OVERSAMPLING
        ConfigDict['RESAMPLING_TYPE'] = '%s' %RESAMPLING_TYPE
        ConfigDict['SUBTRACT_BACK'] = '%s' %SUBTRACT_BACK

        ConfigDict['COMBINE'] = 'Y'  # trivial here
        ConfigDict['COMBINE_TYPE'] = 'MEDIAN'  # trivial here
        
        ConfigDict['WEIGHT_SUFFIX'] = '.weight.fits'  # trivial here
        ConfigDict['WRITE_XML'] = 'N'
        ConfigDict['VERBOSE_TYPE'] = '%s' %VERBOSE_TYPE
        
        swarp_config_path = AMConfig_Maker.AMCM(MDIR=TDIR, \
            AstroMatic_KEY='swarp', ConfigDict=ConfigDict, tag='PYSWarp')

        # * make temporary FITS_resamp & FIT_resamp_weight in TDIR
        fname_obj = pa.basename(FITS_obj)[:-5]
        tFITS_resamp = os.path.join(TDIR, '%s.tmp_resamp.fits' %fname_obj)
        tFITS_resamp_weight = os.path.join(TDIR, '%s.tmp_resamp_weight.fits' %fname_obj)

        # * build .head file from FITS_ref
        phr_ref = fits.getheader(FITS_ref, ext=0)
        w_ref = Read_WCS.RW(phr_ref, VERBOSE_LEVEL=VERBOSE_LEVEL)

        _headfile = tFITS_resamp[:-5] + '.head'
        wcshdr_ref = w_ref.to_header(relax=True)   # SIP distorsion requires relax=True

        wcshdr_ref['BITPIX'] = phr_ref['BITPIX']
        wcshdr_ref['NAXIS'] = phr_ref['NAXIS']

        wcshdr_ref['NAXIS1'] = phr_ref['NAXIS1']
        wcshdr_ref['NAXIS2'] = phr_ref['NAXIS2']

        wcshdr_ref.tofile(_headfile)

        # * trigger SWarp 
        command = "cd %s && swarp %s -IMAGEOUT_NAME %s -WEIGHTOUT_NAME %s -c %s"  \
            %(TDIR, FITS_obj, tFITS_resamp, tFITS_resamp_weight, swarp_config_path)
        print('MeLOn CheckPoint: Run SWarp Command ... \n %s' %command)
        os.system(command)

        # * make a combined header for output FITS
        hdr_op = Combine_Header.CH(hdr_base=phr_obj, hdr_wcs=phr_ref, VERBOSE_LEVEL=VERBOSE_LEVEL)
        
        # * update header by the SWarp generated saturation level
        NEW_SATUR = fits.getheader(tFITS_resamp, ext=0)['SATURATE']
        if SATUR_KEY in hdr_op:
            hdr_op[SATUR_KEY] = (NEW_SATUR, 'MeLOn: PYSWarp')

        # * add history
        hdr_op['SWARP_O'] = (pa.basename(FITS_obj), 'MeLOn: PYSWarp')
        hdr_op['SWARP_R'] = (pa.basename(FITS_ref), 'MeLOn: PYSWarp')
        for key in ConfigDict:
            value = ConfigDict[key]
            pack = ' : '.join(['SWarp Parameter', key, value])
            hdr_op.add_history(pack)

        # * fill the missing data in resampled image [SWarp default 0]
        PixA_resamp, MissingMask = None, None
        try: 
            PixA_resamp = fits.getdata(tFITS_resamp, ext=0).T
            PixA_resamp_weight = fits.getdata(tFITS_resamp_weight, ext=0).T
            MissingMask = PixA_resamp_weight == 0
            if FILL_VALUE is not None:
                PixA_resamp[MissingMask] = FILL_VALUE
            if FITS_resamp is not None:
                hdl_op = fits.HDUList(fits.PrimaryHDU(PixA_resamp.T, header=hdr_op))
                hdl_op.writeto(FITS_resamp, overwrite=True)
        except:
            if VERBOSE_LEVEL in [0, 1, 2]:
                _warn_message = 'SWarp FAILED on [%s]!' %objname
                warnings.warn('MeLOn WARNING: %s' %_warn_message)
        
        os.system('rm -rf %s'%TDIR)

        return PixA_resamp, MissingMask



import cupy as cp

__last_update__ = "2024-09-19"
__author__ = "Shu Liu <shl159@pitt.edu> & Lei Hu <leihu@andrew.cmu.edu>"

class Cupy_WCS_Transform:
    def __init__(self):
        return None

    def read_cd_wcs(self, hdr_wcs, CDKEY="CD"):
        """
        # * Note on the CD matrix transformation:
        The sky coordinate (x, y) relative to reference point can be connected with
        the image coordinate (u, v) relative to reference point by the CD matrix:
            [x]   [CD1_1 CD1_2] [u]
            [y] = [CD2_1 CD2_2] [v]
        where CD1_1, CD1_2, CD2_1, CD2_2 are stored in the FITS header.

        """
        assert hdr_wcs["CTYPE1"] == "RA---TAN"
        assert hdr_wcs["CTYPE2"] == "DEC--TAN"

        N0 = int(hdr_wcs["NAXIS1"])
        N1 = int(hdr_wcs["NAXIS2"])

        CRPIX1 = float(hdr_wcs["CRPIX1"])
        CRPIX2 = float(hdr_wcs["CRPIX2"])

        CRVAL1 = float(hdr_wcs["CRVAL1"])
        CRVAL2 = float(hdr_wcs["CRVAL2"])

        KEYDICT = {
            "N0": N0, "N1": N1,
            "CRPIX1": CRPIX1, "CRPIX2": CRPIX2,
            "CRVAL1": CRVAL1, "CRVAL2": CRVAL2
        }

        CD1_1 = hdr_wcs[f"{CDKEY}1_1"]
        CD1_2 = hdr_wcs[f"{CDKEY}1_2"] if f"{CDKEY}1_2" in hdr_wcs else 0.
        CD2_1 = hdr_wcs[f"{CDKEY}2_1"] if f"{CDKEY}2_1" in hdr_wcs else 0.
        CD2_2 = hdr_wcs[f"{CDKEY}2_2"]

        CD_GPU = cp.array([
            [CD1_1, CD1_2],
            [CD2_1, CD2_2]
        ], dtype=cp.float64)

        return KEYDICT, CD_GPU

    def read_sip_wcs(self, hdr_wcs):
        """
        # * Note on the SIP transformation:
        The image coordinate (u, v) relative to reference point and we have undistorted image coordinate (U, V):
            U = u + f(u, v)
            V = v + g(u, v)
        where f(u, v) and g(u, v) are the SIP distortion functions with polynomial form:
            f(u, v) = sum_{p, q} A_{pq} u^p v^q, p + q <= A_ORDER
            g(u, v) = sum_{p, q} B_{pq} u^p v^q, p + q <= B_ORDER
        These coefficients are stored in the FITS header and define a foward transformation from (u, v) to (U, V).

        The sky coordinate (x, y) relative to reference point can be connected with (U, V) by the reversible CD matrix:
            [x]   [CD1_1 CD1_2] [U]
            [y] = [CD2_1 CD2_2] [V]
        So the forward transformation & CD matrix (with considering the reference point) is equivalent to a pix2world function.

        The reverse question is how to obtain the backward transformation from undistorted (U, V) to distorted (u, v).
        Once we have the backward transformation, we can combine a reversed CD matrix
        (with considering the reference point) to get the world2pix function.

        A simple approach (not accurate) is to assume the forward transformation also has a polynomial form:
            u = U + fp(U, V)
            v = V + gp(U, V)
        where
            fp(U, V) = sum_{p, q} AP_{pq} U^p V^q, p + q <= A_ORDER
            gp(U, V) = sum_{p, q} BP_{pq} U^p V^q, p + q <= B_ORDER

        The coefficients AP_{pq} and BP_{pq} can be obtained by solving the linear equations separately.
            [u - U] = [U^p*V^q] [AP_{pq}]

            shapes:
            b: [u - U]   | (N, 1),
            A: [U^p*V^q] | (N, (A_ORDER+1)*(A_ORDER+2)/2)
            x: [AP_{pq}] | ((A_ORDER+1)*(A_ORDER+2)/2, 1)

            [v - V] = [U^p*V^q] [BP_{pq}]
            b: [v - V]   | (N, 1),
            A: [U^p*V^q] | (N, (B_ORDER+1)*(B_ORDER+2)/2)
            x: [BP_{pq}] | ((B_ORDER+1)*(B_ORDER+2)/2, 1)

            The two matrices encode the "backward" transformation in fp(U, V) and gp(U, V).
            Let's denote the two matrices as FP_UV and GP_UV, respectively.

        # * A WCS example with SIP distortion
        CTYPE1  = 'RA---TAN-SIP'
        CTYPE2  = 'DEC--TAN-SIP'

        CRPIX1  =               2044.0
        CRPIX2  =               2044.0

        CD1_1   = 2.65458074767927E-05
        CD1_2   = -1.2630331175158E-05
        CD2_1   = 1.38917634264203E-05
        CD2_2   = 2.57785917553667E-05

        CRVAL1  =    9.299707734492053
        CRVAL2  =  -43.984663860136145

        A_ORDER =                    4
        A_0_2   =     -4.774423702E-10
        A_0_3   =     -2.462744787E-14
        A_0_4   =       2.28913156E-17
        A_1_1   =      1.076615801E-09
        A_1_2   =     -9.939904553E-14
        A_1_3   =     -5.845336863E-17
        A_2_0   =     -3.717865118E-10
        A_2_1   =     -4.966118494E-15
        A_2_2   =      6.615859199E-17
        A_3_0   =     -3.817793356E-15
        A_3_1   =      3.310020049E-17
        A_4_0   =     -5.166048303E-19

        B_ORDER =                    4
        B_0_2   =      1.504792792E-09
        B_0_3   =     -2.662833066E-15
        B_0_4   =     -8.383877471E-20
        B_1_1   =      1.204553316E-10
        B_1_2   =       6.57748238E-14
        B_1_3   =     -6.998508554E-18
        B_2_0   =      4.136013664E-10
        B_2_1   =      5.339208582E-14
        B_2_2   =      6.063403412E-17
        B_3_0   =      1.486316541E-14
        B_3_1   =     -9.102668806E-17
        B_4_0   =     -5.174323631E-17

        """
        assert hdr_wcs["CTYPE1"] == "RA---TAN-SIP"
        assert hdr_wcs["CTYPE2"] == "DEC--TAN-SIP"

        N0 = int(hdr_wcs["NAXIS1"])
        N1 = int(hdr_wcs["NAXIS2"])

        CRPIX1 = float(hdr_wcs["CRPIX1"])
        CRPIX2 = float(hdr_wcs["CRPIX2"])

        CRVAL1 = float(hdr_wcs["CRVAL1"])
        CRVAL2 = float(hdr_wcs["CRVAL2"])

        A_ORDER = int(hdr_wcs["A_ORDER"])
        B_ORDER = int(hdr_wcs["B_ORDER"])

        KEYDICT = {
            "N0": N0, "N1": N1,
            "CRPIX1": CRPIX1, "CRPIX2": CRPIX2,
            "CRVAL1": CRVAL1, "CRVAL2": CRVAL2,
            "A_ORDER": A_ORDER, "B_ORDER": B_ORDER
        }

        CD_GPU = cp.array([
            [hdr_wcs["CD1_1"], hdr_wcs["CD1_2"]],
            [hdr_wcs["CD2_1"], hdr_wcs["CD2_2"]]
        ], dtype=cp.float64)

        A_SIP_GPU = cp.zeros((A_ORDER+1, A_ORDER+1), dtype=cp.float64)
        B_SIP_GPU = cp.zeros((B_ORDER+1, B_ORDER+1), dtype=cp.float64)

        for p in range(A_ORDER + 1):
            for q in range(0, A_ORDER - p + 1):
                keyword = f"A_{p}_{q}"
                if keyword in hdr_wcs:
                    A_SIP_GPU[p, q] = hdr_wcs[keyword]

        for p in range(B_ORDER + 1):
            for q in range(0, B_ORDER - p + 1):
                keyword = f"B_{p}_{q}"
                if keyword in hdr_wcs:
                    B_SIP_GPU[p, q] = hdr_wcs[keyword]

        return KEYDICT, CD_GPU, A_SIP_GPU, B_SIP_GPU

    def cd_transform(self, IMAGE_X_GPU, IMAGE_Y_GPU, CD_GPU):
        """CD matrix transformation from image to world"""
        WORLD_X_GPU, WORLD_Y_GPU = cp.matmul(CD_GPU, cp.array([IMAGE_X_GPU, IMAGE_Y_GPU]))
        return WORLD_X_GPU, WORLD_Y_GPU

    def cd_transform_inv(self, WORLD_X_GPU, WORLD_Y_GPU, CD_GPU):
        """CD matrix transformation from world to image"""
        # CD_inv_GPU = cp.linalg.inv(CD_GPU)
        CD_inv_GPU = 1. / (CD_GPU[0, 0] * CD_GPU[1, 1] - CD_GPU[0, 1] * CD_GPU[1, 0]) * \
            cp.array([[CD_GPU[1, 1], -CD_GPU[0, 1]], [-CD_GPU[1, 0], CD_GPU[0, 0]]])
        IMAGE_X_GPU, IMAGE_Y_GPU = cp.matmul(CD_inv_GPU, cp.array([WORLD_X_GPU, WORLD_Y_GPU]))
        return IMAGE_X_GPU, IMAGE_Y_GPU

    def sip_forward_transform(self, u_GPU, v_GPU, A_SIP_GPU, B_SIP_GPU):
        """Forward SIP transformation from (u, v) to (U, V)"""
        A_ORDER = A_SIP_GPU.shape[0] - 1
        U_GPU = u_GPU + cp.zeros_like(u_GPU)
        for p in range(A_ORDER + 1):
            for q in range(A_ORDER - p + 1):
                U_GPU += A_SIP_GPU[p, q] * u_GPU**p * v_GPU**q

        B_ORDER = B_SIP_GPU.shape[0] - 1
        V_GPU = v_GPU + cp.zeros_like(v_GPU)
        for p in range(B_ORDER + 1):
            for q in range(B_ORDER - p + 1):
                V_GPU += B_SIP_GPU[p, q] * u_GPU**p * v_GPU**q
        return U_GPU, V_GPU

    def sip_backward_matrix(self, U_GPU, V_GPU, ORDER):
        """Compute the backward matrix P_UV (FP_UV or GP_UV)"""
        Nsamp = len(U_GPU)
        P_UV_GPU = cp.zeros((Nsamp, (ORDER+1)*(ORDER+2)//2), dtype=cp.float64)
        idx = 0
        for p in range(ORDER + 1):
            for q in range(ORDER - p + 1):
                P_UV_GPU[:, idx] = U_GPU**p * V_GPU**q
                idx += 1
        return P_UV_GPU

    def lstsq_sip_backward_transform(self, u_GPU, v_GPU, U_GPU, V_GPU, A_ORDER, B_ORDER):
        """Compute the backward transformation from (U, V) to (u, v)"""
        FP_UV_GPU = self.sip_backward_matrix(U_GPU=U_GPU, V_GPU=V_GPU, ORDER=A_ORDER)
        GP_UV_GPU = self.sip_backward_matrix(U_GPU=U_GPU, V_GPU=V_GPU, ORDER=B_ORDER)

        AP_lstsq_GPU = cp.linalg.lstsq(FP_UV_GPU, (u_GPU - U_GPU).reshape(-1, 1), rcond=None)[0]
        BP_lstsq_GPU = cp.linalg.lstsq(GP_UV_GPU, (v_GPU - V_GPU).reshape(-1, 1), rcond=None)[0]
        return FP_UV_GPU, GP_UV_GPU, AP_lstsq_GPU, BP_lstsq_GPU

    def sip_backward_transform(self, U_GPU, V_GPU, FP_UV_GPU, GP_UV_GPU, AP_GPU, BP_GPU):
        """Backward transformation from (U, V) to (u, v)"""
        u_GPU = U_GPU + cp.matmul(FP_UV_GPU, AP_GPU)[:, 0]
        v_GPU = V_GPU + cp.matmul(GP_UV_GPU, BP_GPU)[:, 0]
        return u_GPU, v_GPU

    def random_coord_sampling(self, N0, N1, CRPIX1, CRPIX2, Nsamp=1024, RANDOM_SEED=10086):
        """Random sampling of coordinates"""
        if RANDOM_SEED is not None:
            cp.random.seed(RANDOM_SEED)
        u_GPU = cp.random.uniform(0.5, N0+0.5, Nsamp, dtype=cp.float64) - CRPIX1
        v_GPU = cp.random.uniform(0.5, N1+0.5, Nsamp, dtype=cp.float64) - CRPIX2
        return u_GPU, v_GPU



# IMPORTS Standard
import cupy as cp
import numpy as np

# IMPORTS Internal
from sfft.PureCupyCustomizedPacket import PureCupy_Customized_Packet
from sfft.utils.PureCupyFFTKits import PureCupy_FFTKits
from sfft.utils.PatternRotationCalculator import PatternRotation_Calculator
from sfft.utils.PureCupyDeCorrelationCalculator import PureCupy_DeCorrelation_Calculator
from sfft.utils.ResampKits import Cupy_ZoomRotate
from sfft.utils.ResampKits import Cupy_Resampling
from sfft.utils.SkyLevelEstimator import SkyLevel_Estimator
from sfft.utils.SFFTSolutionReader import Realize_MatchingKernel

__last_update__ = "2025-06-17"
__author__ = "Lei Hu <leihu@andrew.cmu.edu>"

class SpaceSFFT_CupyFlow:
    """Run A Cupy WorkFlow for SFFT subtraction"""

    def __init__(self, hdr_target, hdr_object, 
                 target_skyrms, object_skyrms, 
                 PixA_target_GPU, PixA_object_GPU,
                 PixA_targetVar_GPU, PixA_objectVar_GPU,
                 PixA_target_DMASK_GPU, PixA_object_DMASK_GPU, 
                 PSF_target_GPU, PSF_object_GPU,
                 sci_is_target=True,
                 GKerHW=9, KerPolyOrder=2, BGPolyOrder=0, ConstPhotRatio=True, 
                 CUDA_DEVICE_4SUBTRACT='0', GAIN=1.0, RANDOM_SEED=10086):
        """Do things.

        Parameters
        ----------
           hdr_target: astropy header
              The target image has the coordinate system to which we are resampling.  This is that image's hader.

           hdr_object: astropy header
              Original (unresampled) header of the image to be resampled to match target.

           target_skyrms: float

           object_skyrms: float

           PixA_target_GPU: cupy array (float64)
              2d image data of target, indexed by x, y.  (Note that raw
              data read from fits files are indexed y, x; transpose to
              get this.)

           PixA_object_GPU: cupy array (float64)
              2d image data of original image, indexed by x, y.

           PixA_targetVar_GPU: cupy array (float64)
              2d image variance of original image, indexed by x, y.

           PixA_objectVar_GPU: cupy array (float64)
              2d image variance of original image, indexed by x, y. 

           PixA_target_DMASK_GPU: cupy array (bool)
              2d detection mask for target image
        
           PixA_object_DMASK_GPU: cupy array (bool)
              2d detection mask for unresampled object image

           PSF_target_GPU: cupy array (float64)
              2d PSF model; axis lengths must be odd.  center of PSF is
              center of center pixel.

           PSF_object_GPU: cupy array (float64)
              object PSF

           sci_is_target : bool
              If True, will subtract object - target.  If false, will subtract target - object.

           GKerHW: int
              Matching kernel half-width (full width is 2*GkerHW + 1 )
           
           KerPolyOrder: int
              Order of spatial variation in kernel

           BGPolyOrder: int
              Order of differential background 2d polynomial.  (Usually
              just leave this 0, we assume the image is sky subtracted.)

           ConstPhotRatio: bool
              Assume relative zeropoints of target and (resampled)
              object have no spatial variation.

           CUDA_DEVICE_4SUBTRACT: str, default '0'
              Which CUDA device to use.

           GAIN: float
              e-/ADU gain for both images.  (So, poisson noise, σ_adu = √(gain) * adu.)

           RANDOM_SEED: int default 10086
              Random seed to use to CR.resamp_projection_sip.  TODO :
              make it so that when this is None, a "real" random seed is
              generated.

        """

        assert PixA_target_GPU.flags['C_CONTIGUOUS']
        assert PixA_object_GPU.flags['C_CONTIGUOUS']
        
        assert PixA_target_DMASK_GPU.flags['C_CONTIGUOUS']
        assert PixA_object_DMASK_GPU.flags['C_CONTIGUOUS']
        
        assert PSF_target_GPU.flags['C_CONTIGUOUS']
        assert PSF_object_GPU.flags['C_CONTIGUOUS']

        self.hdr_target = hdr_target
        self.hdr_object = hdr_object

        self.target_skyrms = target_skyrms
        self.object_skyrms = object_skyrms

        self.PixA_target_GPU = PixA_target_GPU
        self.PixA_object_GPU = PixA_object_GPU

        if PixA_targetVar_GPU.dtype != cp.float64:
            PixA_targetVar_GPU = PixA_targetVar_GPU.astype(cp.float64)
        self.PixA_targetVar_GPU = PixA_targetVar_GPU
        if PixA_objectVar_GPU.dtype != cp.float64:
            PixA_objectVar_GPU = PixA_objectVar_GPU.astype(cp.float64)
        self.PixA_objectVar_GPU = PixA_objectVar_GPU

        if PixA_target_DMASK_GPU.dtype != cp.float64:
            PixA_target_DMASK_GPU = PixA_target_DMASK_GPU.astype(cp.float64)
        self.PixA_target_DMASK_GPU = PixA_target_DMASK_GPU
        if PixA_object_DMASK_GPU.dtype != cp.float64:
            PixA_object_DMASK_GPU = PixA_object_DMASK_GPU.astype(cp.float64)
        self.PixA_object_DMASK_GPU = PixA_object_DMASK_GPU

        self.PSF_target_GPU = PSF_target_GPU
        self.PSF_object_GPU = PSF_object_GPU

        self.sci_is_target = sci_is_target
        self.GKerHW = GKerHW
        self.KerPolyOrder = KerPolyOrder
        self.BGPolyOrder = BGPolyOrder
        self.ConstPhotRatio = ConstPhotRatio
        self.CUDA_DEVICE_4SUBTRACT = CUDA_DEVICE_4SUBTRACT
        self.GAIN = GAIN
        self.RANDOM_SEED = 10086


    def resampling_image_mask_psf( self ):
        # * step 0. run resampling for input object image, variance image, mask, and PSF
        CR = Cupy_Resampling(RESAMP_METHOD="BILINEAR", VERBOSE_LEVEL=1)
        XX_proj_GPU, YY_proj_GPU = CR.resamp_projection_sip(hdr_obj=self.hdr_object,
                                                            hdr_targ=self.hdr_target,
                                                            NSAMP=1024,
                                                            RANDOM_SEED=self.RANDOM_SEED)
        
        # check if projection completely outside of target image
        # TODO: this check is currently not smart...
        NTX = int(self.hdr_target["NAXIS1"])
        NTY = int(self.hdr_target["NAXIS2"])
        NPIX_INNER = cp.sum(cp.logical_and( cp.logical_and(XX_proj_GPU >= 0.5, XX_proj_GPU < NTX+0.5),
                                            cp.logical_and(YY_proj_GPU >= 0.5, YY_proj_GPU < NTY+0.5) ))
        assert NPIX_INNER > 0, "SFFT Error: Projection of object image is completely outside of target image!"

        # Object image:
        PixA_Eobj_GPU, EProjDict = CR.frame_extension(XX_proj_GPU=XX_proj_GPU,
                                                      YY_proj_GPU=YY_proj_GPU, 
                                                      PixA_obj_GPU=self.PixA_object_GPU,
                                                      PAD_FILL_VALUE=0.,
                                                      NAN_FILL_VALUE=0.)

        self.PixA_resamp_object_GPU = CR.resampling(PixA_Eobj_GPU=PixA_Eobj_GPU,
                                                    EProjDict=EProjDict,
                                                    USE_SHARED_MEMORY=False)

        # Variance image:
        PixA_EobjVar_GPU, EProjDict_Var = CR.frame_extension(XX_proj_GPU=XX_proj_GPU,
                                                             YY_proj_GPU=YY_proj_GPU, 
                                                             PixA_obj_GPU=self.PixA_objectVar_GPU,
                                                             PAD_FILL_VALUE=0.,
                                                             NAN_FILL_VALUE=0.)

        self.PixA_resamp_objectVar_GPU = CR.resampling(PixA_Eobj_GPU=PixA_EobjVar_GPU,
                                                       EProjDict=EProjDict,
                                                       USE_SHARED_MEMORY=False)

        # Mask:
        PixA_Eobj_GPU, EProjDict = CR.frame_extension(XX_proj_GPU=XX_proj_GPU,
                                                      YY_proj_GPU=YY_proj_GPU, 
                                                      PixA_obj_GPU=self.PixA_object_DMASK_GPU,
                                                      PAD_FILL_VALUE=0.,
                                                      NAN_FILL_VALUE=0.)

        del XX_proj_GPU
        del YY_proj_GPU
        
        self.PixA_resamp_object_DMASK_GPU = CR.resampling(PixA_Eobj_GPU=PixA_Eobj_GPU,
                                                          EProjDict=EProjDict,
                                                          USE_SHARED_MEMORY=False)
        self.BlankMask_GPU = self.PixA_resamp_object_GPU == 0.


        # PSF:
        PATTERN_ROTATE_ANGLE = PatternRotation_Calculator.PRC(hdr_obj=self.hdr_object, hdr_targ=self.hdr_target)

        self.PSF_resamp_object_GPU = Cupy_ZoomRotate.CZR(PixA_obj_GPU=self.PSF_object_GPU,
                                                         ZOOM_SCALE_X=1.,
                                                         ZOOM_SCALE_Y=1.,
                                                         OUTSIZE_PARIRY_X='UNCHANGED',
                                                         OUTSIZE_PARIRY_Y='UNCHANGED',
                                                         PATTERN_ROTATE_ANGLE=PATTERN_ROTATE_ANGLE,
                                                         RESAMP_METHOD='BILINEAR',
                                                         PAD_FILL_VALUE=0.,
                                                         NAN_FILL_VALUE=0.,
                                                         THREAD_PER_BLOCK=8,
                                                         USE_SHARED_MEMORY=False,
                                                         VERBOSE_LEVEL=2)

    def cross_convolution( self ):
        # * step 1. cross convolution
        self.PixA_Ctarget_GPU = PureCupy_FFTKits.FFT_CONVOLVE(PixA_Inp_GPU=self.PixA_target_GPU,
                                                              KERNEL_GPU=self.PSF_resamp_object_GPU, 
                                                              PAD_FILL_VALUE=0.,
                                                              NAN_FILL_VALUE=None,
                                                              NORMALIZE_KERNEL=True,
                                                              FORCE_OUTPUT_C_CONTIGUOUS=True,
                                                              FFT_BACKEND="Cupy")

        self.PixA_Cresamp_object_GPU = PureCupy_FFTKits.FFT_CONVOLVE(PixA_Inp_GPU=self.PixA_resamp_object_GPU,
                                                                     KERNEL_GPU=self.PSF_target_GPU,
                                                                     PAD_FILL_VALUE=0.,
                                                                     NAN_FILL_VALUE=None,
                                                                     NORMALIZE_KERNEL=True,
                                                                     FORCE_OUTPUT_C_CONTIGUOUS=True,
                                                                     FFT_BACKEND="Cupy")

    def sfft_subtraction( self ):
        # * step 2. sfft subtraction
        LYMASK_BKG_GPU = cp.logical_or(self.PixA_target_DMASK_GPU == 0, self.PixA_resamp_object_DMASK_GPU < 0.1)   # background-mask

        NaNmask_Ctarget_GPU = cp.isnan(self.PixA_Ctarget_GPU)
        NaNmask_Cresamp_object_GPU = cp.isnan(self.PixA_Cresamp_object_GPU)
        if NaNmask_Ctarget_GPU.any() or NaNmask_Cresamp_object_GPU.any():
            NaNmask_GPU = cp.logical_or(NaNmask_Ctarget_GPU, NaNmask_Cresamp_object_GPU)
            ZeroMask_GPU = cp.logical_or(NaNmask_GPU, LYMASK_BKG_GPU)
        else:
            ZeroMask_GPU = LYMASK_BKG_GPU

        del LYMASK_BKG_GPU
            
        PixA_mCtarget_GPU = self.PixA_Ctarget_GPU.copy()
        PixA_mCtarget_GPU[ZeroMask_GPU] = 0.

        PixA_mCresamp_object_GPU = self.PixA_Cresamp_object_GPU.copy()
        PixA_mCresamp_object_GPU[ZeroMask_GPU] = 0.

        del ZeroMask_GPU
        
        # trigger sfft subtraction
        if self.sci_is_target:
            PixA_REF_GPU = self.PixA_Cresamp_object_GPU
            PixA_SCI_GPU = self.PixA_Ctarget_GPU
            PixA_mREF_GPU = PixA_mCresamp_object_GPU
            PixA_mSCI_GPU = PixA_mCtarget_GPU
        else:
            PixA_REF_GPU = self.PixA_Ctarget_GPU
            PixA_SCI_GPU = self.PixA_Cresamp_object_GPU
            PixA_mREF_GPU = PixA_mCtarget_GPU
            PixA_mSCI_GPU = PixA_mCresamp_object_GPU
        
        self.Solution_GPU, self.PixA_DIFF_GPU = PureCupy_Customized_Packet.PCCP(
            PixA_REF_GPU=PixA_REF_GPU,
            PixA_SCI_GPU=PixA_SCI_GPU,
            PixA_mREF_GPU=PixA_mREF_GPU,
            PixA_mSCI_GPU=PixA_mSCI_GPU,
            ForceConv='REF' if self.sci_is_target else 'NEW',
            GKerHW=self.GKerHW,
            KerPolyOrder=self.KerPolyOrder,
            BGPolyOrder=self.BGPolyOrder,
            ConstPhotRatio=self.ConstPhotRatio, 
            CUDA_DEVICE_4SUBTRACT=self.CUDA_DEVICE_4SUBTRACT
        )
        self.PixA_DIFF_GPU[self.BlankMask_GPU] = 0.

    def find_decorrelation( self ):
        # * step 3. perform decorrelation in Fourier domain
        # extract matching kernel at the center
        N0, N1 = self.PixA_DIFF_GPU.shape
        L0, L1 = 2*self.GKerHW + 1, 2*self.GKerHW + 1
        DK = self.KerPolyOrder
        Fpq = int((self.BGPolyOrder+1)*(self.BGPolyOrder+2)/2)
        XY_q = np.array([[N0/2.+0.5, N1/2.+0.5]])

        self.Solution = cp.asnumpy(self.Solution_GPU)
        MATCH_KERNEL_GPU = cp.array(Realize_MatchingKernel(XY_q=XY_q).FromArray(
            Solution=self.Solution, N0=N0, N1=N1, L0=L0, L1=L1, DK=DK, Fpq=Fpq
        )[0], dtype=cp.float64)

        # NOTE -- assuming below that the resampled object image has the same
        #   skyrms as the original object image.  (This is ~OK.)
        self.FKDECO_GPU = PureCupy_DeCorrelation_Calculator.PCDC(NX_IMG=N0,
                                                                 NY_IMG=N1,
                                                                 KERNEL_GPU_JQueue=[self.PSF_target_GPU], 
                                                                 BKGSIG_JQueue=[self.object_skyrms],
                                                                 KERNEL_GPU_IQueue=[self.PSF_resamp_object_GPU],
                                                                 BKGSIG_IQueue=[self.target_skyrms], 
                                                                 MATCH_KERNEL_GPU=MATCH_KERNEL_GPU,
                                                                 REAL_OUTPUT=False,
                                                                 REAL_OUTPUT_SIZE=None, 
                                                                 NORMALIZE_OUTPUT=True,
                                                                 VERBOSE_LEVEL=2)


    def apply_decorrelation( self, img ):
        # do decorrelation

        padxl = 0
        padyl = 0
        padxh = 0
        padyh = 0
        # Implicitly assuming img is smaller here
        if img.shape != self.FKDECO_GPU.shape:
            padx = self.FKDECO_GPU.shape[0] - img.shape[0]
            padxl = padx // 2
            padxh = padx - padxl
            pady = self.FKDECO_GPU.shape[1] - img.shape[1]
            padyl = pady // 2
            padyh = pady - padyl
            img = cp.pad( img, ( (padxl, padxh), (padyl, padyh) ) )
            
        Fdecor = cp.fft.fft2( img )
        decorimg = cp.fft.ifft2( Fdecor * self.FKDECO_GPU ).real

        return decorimg[ padxl:(decorimg.shape[0]-padxh) , padyl:(decorimg.shape[1]-padyh) ]
        
        # FPixA_DIFF_GPU = cp.fft.fft2(self.PixA_DIFF_GPU)
        # self.PixA_DCDIFF_GPU = cp.fft.ifft2(FPixA_DIFF_GPU * FKDECO_GPU).real

    # def noise_decorrelation_snr_estimation( self, bkgsig ):
        # This function is both broken and possibly not necessary since we have the score image. 
        # * step 4. noise decorrelation & SNR estimation 
        # roughly estimate the SNR map for the decorrelated difference image
        # WARNING: the noise propagation is highly simplified.

        # Note: we ignored the background noise change by resampling
        # self.BKGSIG_resamp_object = self.hdr_object['BKG_SIG']
        # self.BKGSIG_target = self.hdr_target['BKG_SIG']

        # PixA_vartarget_GPU = cp.clip(self.PixA_target_GPU/GAIN, a_min=0.0, a_max=None) + self.BKGSIG_target**2
        # PixA_varresamp_object_GPU = cp.clip(self.PixA_resamp_object_GPU/GAIN, a_min=0.0, a_max=None) + self.BKGSIG_resamp_object**2
        # self.PixA_NDIFF_GPU = cp.sqrt(PixA_vartarget_GPU + PixA_varresamp_object_GPU)
        # self.PixA_DSNR_GPU = PixA_DCDIFF_GPU / PixA_NDIFF_GPU

        # return PixA_DIFF_GPU, PixA_DCDIFF_GPU, PixA_DSNR_GPU

    def create_score_image( self ):
        
        # retrieve the decorrelated PSF
        # Note: here we assume the same pixel size for PSF and imgaes.
        NX, NY = self.PixA_target_GPU.shape
        PSF_object_CSZ_GPU = PureCupy_FFTKits.KERNEL_CSZ(KERNEL_GPU=self.PSF_object_GPU, NX_IMG=NX, NY_IMG=NY)
        PSF_target_CSZ_GPU = PureCupy_FFTKits.KERNEL_CSZ(KERNEL_GPU=self.PSF_target_GPU, NX_IMG=NX, NY_IMG=NY)
        FPSF_dDIFF_GPU = cp.fft.fft2(PSF_object_CSZ_GPU) * cp.fft.fft2(PSF_target_CSZ_GPU) * self.FKDECO_GPU

        # apply the decorrelation on difference image again (redundant, a workaround) 
        FPixA_DIFF_GPU = cp.fft.fft2( self.PixA_DIFF_GPU )
        FPixA_dDIFF_GPU = FPixA_DIFF_GPU * self.FKDECO_GPU

        FPixA_SCORE_GPU = FPixA_dDIFF_GPU * cp.conj(FPSF_dDIFF_GPU)
        PixA_SCORE_GPU = cp.fft.ifft2(FPixA_SCORE_GPU).real

        # an ad-hoc correction to make score image has standrd Gaussian distribution at background
        skysig_SCORE = SkyLevel_Estimator.SLE(PixA_obj=cp.asnumpy(PixA_SCORE_GPU))[1]
        PixA_SCORE_GPU /= skysig_SCORE

        return PixA_SCORE_GPU

    def create_variance_image( self ):

        assert self.PixA_targetVar_GPU.flags['C_CONTIGUOUS']
        assert self.PixA_resamp_objectVar_GPU.flags['C_CONTIGUOUS']

        # calculate variance image for (un-decorrelated) difference image
        NX, NY = self.PixA_target_GPU.shape
        PSF_object_CSZ_GPU = PureCupy_FFTKits.KERNEL_CSZ(KERNEL_GPU=self.PSF_object_GPU, NX_IMG=NX, NY_IMG=NY)
        PSF_target_CSZ_GPU = PureCupy_FFTKits.KERNEL_CSZ(KERNEL_GPU=self.PSF_target_GPU, NX_IMG=NX, NY_IMG=NY)

        # Note: convolve a squared kernel on the variance image to get the variance of the convolved image
        # Note: let's skip the matching kernel here, as it is expected to be a minor compensation.
        FPixA_DIFFVar_GPU = cp.fft.fft2(self.PixA_resamp_objectVar_GPU) * cp.fft.fft2(PSF_target_CSZ_GPU ** 2) + \
            cp.fft.fft2(self.PixA_targetVar_GPU) * cp.fft.fft2(PSF_object_CSZ_GPU ** 2)
        
        FPixA_dDIFFVar_GPU = FPixA_DIFFVar_GPU * cp.fft.fft2(cp.fft.ifft2(self.FKDECO_GPU) ** 2)
        PixA_dDIFFVar_GPU = cp.fft.ifft2(FPixA_dDIFFVar_GPU).real
        
        return PixA_dDIFFVar_GPU

    # Do we need this?  We should just unreference the object
    def cleanup( self ):
        pass
        # TODO
        # del self.hdr_target
        # del self.hdr_object
        # del self.PixA_target_GPU
        # del self.PixA_object_GPU
        # del self.PixA_target_DMASK_GPU
        # del self.PixA_object_DMASK_GPU
        # del self.PSF_target_GPU
        # del self.PSF_object_GPU
        # del self.GKerHW
        # del self.KerPolyOrder
        # del self.BGPolyOrder
        # del self.ConstPhotRatio
        # del self.CUDA_DEVICE_4SUBTRACT
        # del self.GAIN
        # del self.which_convolve
        # del self.RANDOM_SEED
        # del self.PixA_resamp_object_GPU        
        # del self.PixA_resamp_object_DMASK_GPU
        # del self.BlankMask_GPU
        # del self.PSF_resamp_object_GPU
        # del self.PixA_Ctarget_GPU
        # del self.PixA_Cresamp_object_GPU

# class SpaceSFFT_CupyFlow_NVTX:
#     @staticmethod
#     def SSCFN(hdr_target, hdr_object, PixA_target_GPU, PixA_object_GPU, PixA_target_DMASK_GPU, PixA_object_DMASK_GPU, 
#         PSF_target_GPU, PSF_object_GPU, GKerHW=9, KerPolyOrder=2, BGPolyOrder=0, ConstPhotRatio=True, 
#         CUDA_DEVICE_4SUBTRACT='0', GAIN=1.0):
#         """Run A Cupy WorkFlow for SFFT subtraction"""

#         assert PixA_target_GPU.flags['C_CONTIGUOUS']
#         assert PixA_object_GPU.flags['C_CONTIGUOUS']
        
#         assert PixA_target_DMASK_GPU.flags['C_CONTIGUOUS']
#         assert PixA_object_DMASK_GPU.flags['C_CONTIGUOUS']
        
#         assert PSF_target_GPU.flags['C_CONTIGUOUS']
#         assert PSF_object_GPU.flags['C_CONTIGUOUS']

#         assert "BKG_SIG" in hdr_target
#         assert "BKG_SIG" in hdr_object

#         import nvtx
        
#         # * step 0. run resampling for input science image, mask and PSF
#         with nvtx.annotate("0-resamp", color="#FFD900"):
#             CR = Cupy_Resampling(RESAMP_METHOD="BILINEAR", VERBOSE_LEVEL=1)
#             XX_proj_GPU, YY_proj_GPU = CR.resamp_projection_sip(hdr_obj=hdr_object, hdr_targ=hdr_target, NSAMP=1024, RANDOM_SEED=10086)

#             PixA_Eobj_GPU, EProjDict = CR.frame_extension(XX_proj_GPU=XX_proj_GPU, YY_proj_GPU=YY_proj_GPU, 
#                 PixA_obj_GPU=PixA_object_GPU, PAD_FILL_VALUE=0., NAN_FILL_VALUE=0.)
#             PixA_resamp_object_GPU = CR.resampling(PixA_Eobj_GPU=PixA_Eobj_GPU, EProjDict=EProjDict, USE_SHARED_MEMORY=False)

#             PixA_Eobj_GPU, EProjDict = CR.frame_extension(XX_proj_GPU=XX_proj_GPU, YY_proj_GPU=YY_proj_GPU, 
#                 PixA_obj_GPU=PixA_object_DMASK_GPU, PAD_FILL_VALUE=0., NAN_FILL_VALUE=0.)
#             PixA_resamp_object_DMASK_GPU = CR.resampling(PixA_Eobj_GPU=PixA_Eobj_GPU, EProjDict=EProjDict, USE_SHARED_MEMORY=False)
#             BlankMask_GPU = PixA_resamp_object_GPU == 0.

#             PATTERN_ROTATE_ANGLE = PatternRotation_Calculator.PRC(hdr_obj=hdr_object, hdr_targ=hdr_target)
#             PSF_resamp_object_GPU = Cupy_ZoomRotate.CZR(PixA_obj_GPU=PSF_object_GPU, ZOOM_SCALE_X=1., ZOOM_SCALE_Y=1., \
#                 OUTSIZE_PARIRY_X='UNCHANGED', OUTSIZE_PARIRY_Y='UNCHANGED', PATTERN_ROTATE_ANGLE=PATTERN_ROTATE_ANGLE, \
#                 RESAMP_METHOD='BILINEAR', PAD_FILL_VALUE=0., NAN_FILL_VALUE=0., THREAD_PER_BLOCK=8, \
#                 USE_SHARED_MEMORY=False, VERBOSE_LEVEL=2)
        
#         # * step 1. cross convolution
#         with nvtx.annotate("1-crossConv", color="#C51B8A"):
#             PixA_Ctarget_GPU = PureCupy_FFTKits.FFT_CONVOLVE(PixA_Inp_GPU=PixA_target_GPU, KERNEL_GPU=PSF_resamp_object_GPU, 
#                 PAD_FILL_VALUE=0., NAN_FILL_VALUE=None, NORMALIZE_KERNEL=True, FORCE_OUTPUT_C_CONTIGUOUS=True, FFT_BACKEND="Cupy")

#             PixA_Cresamp_object_GPU = PureCupy_FFTKits.FFT_CONVOLVE(PixA_Inp_GPU=PixA_resamp_object_GPU, KERNEL_GPU=PSF_target_GPU,
#                 PAD_FILL_VALUE=0., NAN_FILL_VALUE=None, NORMALIZE_KERNEL=True, FORCE_OUTPUT_C_CONTIGUOUS=True, FFT_BACKEND="Cupy")

#         # * step 2. sfft subtraction
#         with nvtx.annotate("2-sfft", color="#233ED9"):
#             LYMASK_BKG_GPU = cp.logical_or(PixA_target_DMASK_GPU == 0, PixA_resamp_object_DMASK_GPU < 0.1)   # background-mask

#             NaNmask_Ctarget_GPU = cp.isnan(PixA_Ctarget_GPU)
#             NaNmask_Cresamp_object_GPU = cp.isnan(PixA_Cresamp_object_GPU)
#             if NaNmask_Ctarget_GPU.any() or NaNmask_Cresamp_object_GPU.any():
#                 NaNmask_GPU = cp.logical_or(NaNmask_Ctarget_GPU, NaNmask_Cresamp_object_GPU)
#                 ZeroMask_GPU = cp.logical_or(NaNmask_GPU, LYMASK_BKG_GPU)
#             else:
#                 ZeroMask_GPU = LYMASK_BKG_GPU

#             PixA_mCtarget_GPU = PixA_Ctarget_GPU.copy()
#             PixA_mCtarget_GPU[ZeroMask_GPU] = 0.

#             PixA_mCresamp_object_GPU = PixA_Cresamp_object_GPU.copy()
#             PixA_mCresamp_object_GPU[ZeroMask_GPU] = 0.

#             # trigger sfft subtraction
#             ForceConv = 'target'  # force to target
#             Solution_GPU, PixA_DIFF_GPU = PureCupy_Customized_Packet.PCCP(
#                 PixA_target_GPU=PixA_Ctarget_GPU, PixA_resamp_object_GPU=PixA_Cresamp_object_GPU, PixA_mtarget_GPU=PixA_mCtarget_GPU, PixA_mresamp_object_GPU=PixA_mCresamp_object_GPU, 
#                 ForceConv=ForceConv, GKerHW=GKerHW, KerPolyOrder=KerPolyOrder, BGPolyOrder=BGPolyOrder, ConstPhotRatio=ConstPhotRatio, 
#                 CUDA_DEVICE_4SUBTRACT=CUDA_DEVICE_4SUBTRACT
#             )
#             PixA_DIFF_GPU[BlankMask_GPU] = 0.

#         # * step 3. perform decorrelation in Fourier domain
#         with nvtx.annotate("3-decorr", color="#65BD63"):
#             # extract matching kernel at the center
#             N0, N1 = PixA_DIFF_GPU.shape
#             L0, L1 = 2*GKerHW + 1, 2*GKerHW + 1
#             DK = KerPolyOrder
#             Fpq = int((BGPolyOrder+1)*(BGPolyOrder+2)/2)
#             XY_q = np.array([[N0/2.+0.5, N1/2.+0.5]])

#             Solution = cp.asnumpy(Solution_GPU)
#             MATCH_KERNEL_GPU = cp.array(Realize_MatchingKernel(XY_q=XY_q).FromArray(
#                 Solution=Solution, N0=N0, N1=N1, L0=L0, L1=L1, DK=DK, Fpq=Fpq
#             )[0], dtype=cp.float64)

#             # do decorrelation
#             # Note: we ignored the background noise change by resampling
#             BKGSIG_resamp_object = hdr_object['BKG_SIG']  
#             BKGSIG_target = hdr_target['BKG_SIG']

#             FKDECO_GPU = PureCupy_DeCorrelation_Calculator.PCDC(NX_IMG=N0, NY_IMG=N1, KERNEL_GPU_JQueue=[PSF_target_GPU], 
#                 BKGSIG_JQueue=[BKGSIG_resamp_object], KERNEL_GPU_IQueue=[PSF_resamp_object_GPU], BKGSIG_IQueue=[BKGSIG_target], 
#                 MATCH_KERNEL_GPU=MATCH_KERNEL_GPU, REAL_OUTPUT=False, REAL_OUTPUT_SIZE=None, 
#                 NORMALIZE_OUTPUT=True, VERBOSE_LEVEL=2)

#             FPixA_DIFF_GPU = cp.fft.fft2(PixA_DIFF_GPU)
#             PixA_DCDIFF_GPU = cp.fft.ifft2(FPixA_DIFF_GPU * FKDECO_GPU).real
        
#         # * step 4. noise decorrelation & SNR estimation 
#         with nvtx.annotate("4-dsnr", color="#00D0FF"):
#             # roughly estimate the SNR map for the decorrelated difference image
#             # WARNING: the noise propagation is highly simplified.

#             PixA_vartarget_GPU = cp.clip(PixA_target_GPU/GAIN, a_min=0.0, a_max=None) + BKGSIG_target**2
#             PixA_varresamp_object_GPU = cp.clip(PixA_resamp_object_GPU/GAIN, a_min=0.0, a_max=None) + BKGSIG_resamp_object**2
#             PixA_NDIFF_GPU = cp.sqrt(PixA_vartarget_GPU + PixA_varresamp_object_GPU)
#             PixA_DSNR_GPU = PixA_DCDIFF_GPU / PixA_NDIFF_GPU

#         return PixA_DIFF_GPU, PixA_DCDIFF_GPU, PixA_DSNR_GPU



import sys
import numpy as np
import os.path as pa
from astropy.io import fits
from astropy.nddata.utils import Cutout2D
from sfft.utils.ReadWCS import Read_WCS
# version: Feb 6, 2023

# improved by Lauren Aldoroty (Duke Univ.)
__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class Stamp_Generator:
    @staticmethod
    def SG(FITS_obj=None, EXTINDEX=0, EXTNAME=None, COORD=None, COORD_TYPE='IMAGE', \
        STAMP_IMGSIZE=None, FILL_VALUE=np.nan, FITS_StpLst=None, VERBOSE_LEVEL=2):

        """
        # MeLOn Notes
        # @ Stamp Generator
        # * Remarks on the 2 basic coordinate systems
        #   A. Fortran & FITS ( ds9 & SEx & WCS ) version
        #      @ Matrix Index start from 1. Thus the first pixel of a matrix is < r_F, c_F > = < 1, 1 >
        #      @ [Basic Consistence Rule in Fortran]
        #        Corresponding xy-Coordinate System: Pixel < r_F, c_F > has Center with xy-coordinate ( x_F, y_F ) = ( r_F, c_F )
        #        NOTE the origin point this system is located at the center point of the pixel < r_F, c_F > = < 0, 0 >, 
        #             which is not physical existed.
        #      @ [Basic inclusion Rule in Fortran]
        #        Given Coordinate is enclosed in which pixel ?
        #        < r_F, c_F > = < int(x_F + 0.5), int(y_F + 0.5) >
        #
        #   B. Numpy & C version
        #      @ Matrix Index start from 0. Thus the first pixel of a matrix is < r_C, c_C > = < 0, 0 >
        #      @ [Basic Consistence Rule in C]
        #        Corresponding xy-Coordinate System: Pixel < r_C, c_C > has Center with xy-coordinate ( x_C, y_C ) = ( r_C, c_C )
        #        NOTE the origin point this system is located at the center point of the pixel < r_C, c_C > = < 0, 0 >, 
        #             which is just the first pixel.
        #      @ [Basic inclusion Rule in C]
        #        Given Coordinate is enclosed in which pixel ?
        #        Hold Pixel < r_C, c_C > = < int(x_C + 0.5), int(y_C + 0.5) >
        #        (Hold Pixel has solid boundary at bottom and left side, and dashed boundary at upper and right side.)
        #   
        #   P.S. < r_F, c_F > = < r_C + 1, c_C + 1 > and < x_F, y_F > = < x_C + 1, y_C + 1 >
        #
        # * Our MeLOn Convention
        #   In our context, we always use C Matrix Index System with Fortran Coordinate System!
        #   henceforth, Let r, c means r_C, c_C and x, y indicate x_F, y_F.
        #
        #   @ [Basic Consistence Rule in Our Convention]
        #     Pixel < r, c > has Center with xy-coordinate (x, y) = (r + 1.0, c + 1.0)      
        #   @ [Basic Inclusion Rule in Our Convention]
        #     Hold Pixel < r, c > = < int(x - 0.5), int(y - 0.5) >
        #   @ Image Center has xy-coordinate (x, y) = (NX/2 + 0.5, NY/2 + 0.5)
        #   @ Image valid domain can be defined as,
        #     Without boundary: 0 <= r, c < NX, NY
        #                       0.5 <= x, y < NX + 0.5, NY + 0.5 
        #     With boundary: NBX, NBY <= r, c < NX - NBX, NY - NBY  
        #                    NBX + 0.5, NBY + 0.5 <= x, y < NX - NBX + 0.5, NY - NBY + 0.5
        #
        # * The convertion function in Astropy
        #   Both w.all_pix2world and w.all_world2pix has an argument called origin 
        #   origin = 1, Fortran xy-coordinate   |   This is what we always use!
        #   origin = 0, C xy-coordinate
        #
        # * Stamp-Procedure for digital image
        #   Given (x, y) ----- Enclosed in Pixel < r, c > = < ⌊x - 0.5⌋,  ⌊y - 0.5⌋ >
        #   Stamp(PixA, x, y, P, Q) = PixA[r - ⌊P/2⌋: r + ⌈P/2⌉, c - ⌊Q/2⌋: c + ⌈Q/2⌉]
        #
        # * Equivalent Stamp-Procedure for given row & column range
        #   Given RowRange = (r0, r1) and ColRange = (c0, c1)
        #   Let P = r1 - r0 and Q = c1 - c0, then 
        #   PixA[r0: r1, c0: c1] = Stamp(PixA, x, y, P, Q) 
        #   where x = r0 + ⌊P/2⌋ + 1.0 and y = c0 + ⌊Q/2⌋ + 1.0
        #
        # * Might be right, but haven't been Checked yet.
        #   PixA[E: -E, F: -F] = Stamp(PixA, U/2, V/2, U-2*E, V-2*F), where U, V = PixA.shape
        #   Stamp(PixA, x, y, P, Q)[E: -E, F: -F] = Stamp(PixA, x, y, P-E, Q-F)
        #
        # * Additional Remarks 
        #   a. Only Support Single-Extension FITS file currently.
        #   b. The Boundary Pixels could be filled with np.nan | median vaule | given value 
        #   c. Here we just care the essential cases with COORD_TYPE='IMAGE' since other 'WORLD' cases could be transformed.
        #   d. Cutout2D size parameter has the form of (ny, nx), however, the required stamp with STAMP_IMGSIZE 
        #      has a form of (NSX, NSY). We require the input of Cutout2D should only be < r, c >, even this 
        #      function actually also support float values as inputs.
        #   e. We should note NAXIS in the header would be automatic updated to adjust the image size.
        #   f. For general purpose, We do use the astropy generated wcs for the stamps.
        #      Alternatively, a more naive method is employed with correction of CRPIX1 & CRPIX2
        #      which is reasonable at least for TPV and SIP transformation, where the equations 
        #      are only dependent on the relative pixel coordinate, that it, the pixel-vector.   
        #      Ref https://fits.gsfc.nasa.gov/registry/tpvwcs/tpv.html  
        #      Ref https://fits.gsfc.nasa.gov/registry/sip/SIP_distortion_v1_0.pdf
        #
        """

        NSX, NSY = STAMP_IMGSIZE
        assert len(COORD.shape) == 2 and COORD.shape[1] == 2
        NPOS = COORD.shape[0]
        
        if EXTNAME is not None:
            hdr_obj = fits.getheader(FITS_obj, extname=EXTNAME)
            data_obj = fits.getdata(FITS_obj, extname=EXTNAME)
        else: 
            hdr_obj = fits.getheader(FITS_obj, ext=EXTINDEX)
            data_obj = fits.getdata(FITS_obj, ext=EXTINDEX)
        dtype_obj = data_obj.dtype
        
        # * read positions
        if np.issubdtype(COORD.dtype, np.floating):
            UCOORD = COORD
        elif np.issubdtype(COORD.dtype, np.integer):
            UCOORD = COORD.astype(float)
            if VERBOSE_LEVEL in [2]:
                _warn_message = 'COORD has integer data type, covert to float automatically!'
                print('MeLOn WARNING: %s' %_warn_message)
        else: 
            _error_message = 'WRONG data type for COORD!'
            sys.exit('MeLOn ERROR: %s' %_error_message)
        
        assert COORD_TYPE in ['IMAGE', 'WORLD']
        if COORD_TYPE == 'IMAGE':
            positions = (UCOORD - 0.5).astype(int)
        if COORD_TYPE == 'WORLD':
            w_obj = Read_WCS.RW(hdr_obj, VERBOSE_LEVEL=VERBOSE_LEVEL)
            positions = (w_obj.all_world2pix(UCOORD, 1) - 0.5).astype(int)

        # * read fill_value, as same type with data
        if type(FILL_VALUE) == float:
            fv = FILL_VALUE
            if np.issubdtype(dtype_obj, np.integer):
                fv = np.array([fv], dtype=dtype_obj)[0]
                if VERBOSE_LEVEL in [1, 2]:
                    _warn_message = 'float FILL_VALUE is coverted to integer automatically!'
                    print('MeLOn WARNING: %s' %_warn_message)
        
        elif type(FILL_VALUE) == int:
            fv = FILL_VALUE
            if np.issubdtype(dtype_obj, np.floating):
                fv = np.array([fv], dtype=dtype_obj)[0]
                if VERBOSE_LEVEL in [2]:
                    _warn_message = 'integer FILL_VALUE is coverted to float automatically!'
                    print('MeLOn WARNING: %s' %_warn_message)
        
        elif FILL_VALUE in ['MEDIAN', 'MEAN']:
            if FILL_VALUE == 'MEDIAN': fv = np.nanmedian(data_obj)
            if FILL_VALUE == 'MEAN': fv = np.nanmean(data_obj)

            if np.issubdtype(dtype_obj, np.integer) and type(fv) == float:
                fv = np.array([fv], dtype=dtype_obj)[0]
                if VERBOSE_LEVEL in [1, 2]:
                    _warn_message = 'float FILL_VALUE is coverted to integer automatically!'
                    print('MeLOn WARNING: %s' %_warn_message)
            
            if np.issubdtype(dtype_obj, np.floating) and type(fv) == int:
                fv = np.array([fv], dtype=dtype_obj)[0]
                if VERBOSE_LEVEL in [2]:
                    _warn_message = 'integer FILL_VALUE is coverted to float automatically!'
                    print('MeLOn WARNING: %s' %_warn_message)

        else:
            _error_message = 'FILL_VALUE is not recognized!'
            sys.exit('MeLOn ERROR: %s' %_error_message)
        
        # * create stamps with astropy function Cutout2D
        PixA_StpLst = []
        for i in range(NPOS):
            try:
                PixA_Stp = Cutout2D(data_obj, position=positions[i], size=(NSY, NSX), \
                    mode='partial', fill_value=fv).data.T
            except: PixA_Stp = fv * np.array(np.ones(STAMP_IMGSIZE), dtype=dtype_obj)
            PixA_StpLst.append(PixA_Stp)

        # * save stamps to FITS files with WCS updated
        if FITS_StpLst is not None:
            hdr_StpLst = [] 
            FNAME = pa.basename(FITS_obj)

            for i in range(NPOS):
                hdr_Stp = hdr_obj.copy()
                row_StpCent, col_StpCent = positions[i]
                row_StpO, col_StpO = row_StpCent - int(NSX/2), col_StpCent - int(NSY/2)
                try:
                    hdr_Stp['CRPIX1'] = float(hdr_Stp['CRPIX1']) - row_StpO
                    hdr_Stp['CRPIX2'] = float(hdr_Stp['CRPIX2']) - col_StpO
                except: pass
                hdr_Stp['COMMENT'] = 'Stamp created from %s ' %FNAME + \
                    'centred at row-colum [%d, %d] ' %(row_StpCent, col_StpCent) + \
                    'with stamp image size (%d %d)' %(NSX, NSY)
                hdr_StpLst.append(hdr_Stp)
            
            for i in range(NPOS):
                PixA_Stp, hdr_Stp = PixA_StpLst[i], hdr_StpLst[i]
                if isinstance(FITS_StpLst, list):
                    FITS_Stp = FITS_StpLst[i]
                elif isinstance(FITS_StpLst, str):
                    FITS_Stp = FITS_StpLst
                fits.HDUList([fits.PrimaryHDU(PixA_Stp.T, header=hdr_Stp)]).writeto(FITS_Stp, overwrite=True)

        return PixA_StpLst



import sys
import time
import warnings
import threading
import numpy as np
import os.path as pa
from astropy.io import fits
import scipy.ndimage as ndimage
from astropy.table import Column
from sfft.AutoSparsePrep import Auto_SparsePrep
from sfft.utils.meta.TimeoutKit import TimeoutAfter
from sfft.utils.SFFTSolutionReader import Realize_FluxScaling
from sfft.sfftcore.SFFTSubtract import GeneralSFFTSubtract
from sfft.sfftcore.SFFTConfigure import SingleSFFTConfigure
# version: Mar 18, 2023

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class MultiEasy_SparsePacket:
    def __init__(self, FITS_REF_Queue, FITS_SCI_Queue, FITS_DIFF_Queue=[], FITS_Solution_Queue=[], \
        ForceConv_Queue=[], GKerHW_Queue=[], KerHWRatio=2.0, KerHWLimit=(2, 20), KerPolyOrder=2, BGPolyOrder=0, \
        ConstPhotRatio=True, MaskSatContam=False, GAIN_KEY='GAIN', SATUR_KEY='ESATUR', BACK_TYPE='MANUAL', \
        BACK_VALUE=0.0, BACK_SIZE=64, BACK_FILTERSIZE=3, DETECT_THRESH=2.0, ANALYSIS_THRESH=2.0, DETECT_MINAREA=5, \
        DETECT_MAXAREA=0, DEBLEND_MINCONT=0.005, BACKPHOTO_TYPE='LOCAL', ONLY_FLAGS=[0], BoundarySIZE=30, \
        XY_PriorSelect_Queue=[], Hough_MINFR=0.1, Hough_PeakClip=0.7, BeltHW=0.2, PointSource_MINELLIP=0.3, \
        MatchTol=None, MatchTolFactor=3.0, COARSE_VAR_REJECTION=True, CVREJ_MAGD_THRESH=0.12, \
        ELABO_VAR_REJECTION=True, EVREJ_RATIO_THREH=5.0, EVREJ_SAFE_MAGDEV=0.04, StarExt_iter=4, \
        XY_PriorBan_Queue=[], PostAnomalyCheck=False, PAC_RATIO_THRESH=5.0, CLEAN_GPU_MEMORY=False, VERBOSE_LEVEL=2):

        """
        # NOTE: This function is to perform Sparse-Flavor SFFT for multiple tasks:
        #       @ this module (only support Cupy backend) would help making the 
        #         best use of the available CPU and GPU resources.

        * Parameters for Sparse-Flavor SFFT [multiple tasks]

        # ----------------------------- Computing Enviornment [Cupy backend ONLY] --------------------------------- #

        -NUM_THREADS_4PREPROC [8]           # the number of Python threads for preprocessing.
                                            # WARNING: Empirically, the current code is only optimized for 
                                            #          NUM_TASK / NUM_THREADS_4PREPROC in [1.0 - 4.0].
                                            #          The computing effeciency goes down sharply as the ratio increases.
                                            #          Too large number of tasks may even cause memory overflow.

        -NUM_THREADS_4SUBTRACT [1]          # the number of Python threads for sfft subtraction.
                                            # says, if you have four GPU cards, then set it to 4.
                                            # NOTE: please do not confused the Python threads 
                                            #       with the CUDA threads (block, threads ...).

        -CUDA_DEVICES_4SUBTRACT [['0']]     # it specifies certain GPU devices (indices) to conduct the subtractions.
                                            # the GPU devices are usually numbered 0 to N-1.
                                            # (you may use command nvidia-smi to check for Nvidia GPU devices).

        -TIMEOUT_4PREPROC_EACHTASK [300]    # timeout of each task during preprocessing.
                                            
        -TIMEOUT_4SUBTRACT_EACHTASK [300]   # timeout of each task during sfft subtraction.

        # ----------------------------- Preprocessing with Source Selection for Image-Masking --------------------------------- #

        # > Configurations for SExtractor

        -GAIN_KEY ['GAIN']                  # SExtractor Parameter GAIN_KEY
                                            # i.e., keyword of GAIN in FITS header (of reference & science)

        -SATUR_KEY ['ESATUR']               # SExtractor Parameter SATUR_KEY
                                            # i.e., keyword of effective saturation in FITS header (of reference & science)
                                            # Remarks: one may think 'SATURATE' is a more common keyword name for saturation level.
                                            #          However, note that Sparse-Flavor SFFT requires sky-subtracted images as inputs, 
                                            #          we need to use the 'effective' saturation level after the sky-subtraction.
                                            #          e.g., set ESATURA = SATURATE - (SKY + 10*SKYSIG)

        -BACK_TYPE ['MANUAL']               # SExtractor Parameter BACK_TYPE = [AUTO or MANUAL].
                                            # As Sparse-Flavor-SFFT requires the input images being sky-subtracted,
                                            # the default setting uses zero-background (i.e., BACK_TYPE='MANUAL' & BACK_VALUE=0.0).
                                            # However, one may also use BACK_TYPE='AUTO' for some specific cases
                                            # E.g., point sources located at the outskirts of a bright galaxy will have a relatively
                                            #       high local background. Even if the image has been sky-subtracted properly, 
                                            #       SExtractor can only detect it when you set BACK_TYPE='AUTO'.
         
        -BACK_VALUE [0.0]                   # SExtractor Parameter BACK_VALUE (only work for BACK_TYPE='MANUAL')

        -BACK_SIZE [64]                     # SExtractor Parameter BACK_SIZE

        -BACK_FILTERSIZE [3]                # SExtractor Parameter BACK_FILTERSIZE

        -DETECT_THRESH [2.0]                # SExtractor Parameter DETECT_THRESH
                                            # NOTE: One may notice that the default DETECT_THRESH in SExtractor is 1.5,
                                            #       Using a 'cold' detection threshold here is to speed up SExtractor.
                                            #       Although DETECT_THRESH = 2.0 means we will miss the faint-end sources with 
                                            #       SNR < 9 (approximately), the cost is generally acceptable for source selection.

        -ANALYSIS_THRESH [2.0]              # SExtractor Parameter ANALYSIS_THRESH
                                            # NOTE: By default, let -ANALYSIS_THRESH = -DETECT_THRESH

        -DETECT_MINAREA [5]                 # SExtractor Parameter DETECT_MINAREA
        
        -DETECT_MAXAREA [0]                 # SExtractor Parameter DETECT_MAXAREA

        -DEBLEND_MINCONT [0.005]            # SExtractor Parameter DEBLEND_MINCONT (typically, 0.001 - 0.005)

        -BACKPHOTO_TYPE ['LOCAL']           # SExtractor Parameter BACKPHOTO_TYPE

        -ONLY_FLAGS [0]                     # Restrict SExtractor Output Photometry Catalog by Source FLAGS
                                            # Common FLAGS (Here None means no restrictions on FLAGS):
                                            # 1: aperture photometry is likely to be biased by neighboring sources 
                                            #    or by more than 10% of bad pixels in any aperture
                                            # 2: the object has been deblended
                                            # 4: at least one object pixel is saturated

        -BoundarySIZE [30]                  # Restrict SExtractor Output Photometry Catalog by Dropping Sources at Boundary 
                                            # NOTE: This would help to avoid selecting sources too close to image boundary. 

        # > Configurations for Source-Selction

        -XY_PriorSelect_Queue [[]]      # A queue of -XY_PriorSelect in sfft.Easy_SparsePacket.
                                        # Recall -XY_PriorSelect: 
                                        # a Numpy array of pixels coordinates, 
                                        # with shape (N, 2) (e.g., [[x0, y0], [x1, y1], ...]).
                                        # this allows sfft to us a prior source selection to solve the subtraction.
                                        # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [SEMI-AUTO] MODE

        -Hough_MINFR [0.1]              # The lower bound of FLUX_RATIO for line feature detection using Hough transformation.
                                        # Setting a proper lower bound can avoid to detect some line features by chance,
                                        # which are not contributed from point sources but resides in the small-FLUX_RATIO region.
                                        # NOTE: Recommended values of Hough_MINFR: 0.1 ~ 1.0
                                        # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE

        -Hough_PeakClip [0.7]           # It determines the lower bound of the sensitivity of the line feature detection.
                                        # NOTE: When the point-source-belt is not very pronounced (e.g., in galaxy dominated fields),
                                        #       one may consider to reduce the parameter from default 0.7 to, says, ~ 0.4.
                                        # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE

        -BeltHW [0.2]                   # The half-width of point-source-belt detected by Hough Transformation.
                                        # Remarks: if you want to tune this parameter, it is helpful to draw 
                                        #          a figure of MAG_AUTO against FLUX_RADIUS.
                                        # NOTE: IT WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE :::: determine point-source-belt

        -PointSource_MINELLIP [0.3]     # An additiona Restriction on ELLIPTICITY (ELLIPTICITY < PointSource_MINELLIP) for point sources.
                                        # NOTE: IT WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE :::: determine point-sources
                                    
        -MatchTol [None]                # Given separation tolerance (pix) for source matching 
                                        # NOTE: IT WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE :::: Cross-Match between REF & SCI
                                        # NOTE: IT WORKS FOR Auto-Sparse-Prep [SEMI-AUTO] MODE :::: Cross-Match between REF & SCI AND
                                        #       Cross-Match between the prior selection and mean image coordinates of matched REF-SCI sources.

        -MatchTolFactor [3.0]           # The separation tolerance (pix) for source matching calculated by FWHM,
                                        # MatchTol = np.sqrt((FWHM_REF/MatchTolFactor)**2 + (FWHM_SCI/MatchTolFactor)**2)
                                        # @ Given precise WCS, one can use a high MatchTolFactor ~3.0
                                        # @ For very sparse fields where WCS can be inaccurate, 
                                        #   one can loosen the tolerance with a low MatchTolFactor ~1.0
                                        # NOTE: IT WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE :::: Cross-Match between REF & SCI
                                        # NOTE: IT WORKS FOR Auto-Sparse-Prep [SEMI-AUTO] MODE :::: Cross-Match between REF & SCI AND
                                        #       Cross-Match between the prior selection and mean image coordinates of matched REF-SCI sources.

        -COARSE_VAR_REJECTION [True]    # Activate Coarse Variable Rejection (CVREJ) or not.
                                        # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE 

        -CVREJ_MAGD_THRESH [0.12]       # CVREJ kicks out the obvious variables by the difference of 
                                        # instrument magnitudes (MAG_AUTO) measured on reference and science, i.e., 
                                        # MAG_AUTO_SCI - MAG_AUTO_REF. A source with difference highly deviated from
                                        # the median level of the field stars will be rejected from the source selection.
                                        # -CVREJ_MAGD_THRESH is the deviation magnitude threshold. 
                                        # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE 
        
        -ELABO_VAR_REJECTION [True]     # WARNING: EVREJ REQUIRES CORRECT GAIN IN FITS HEADER!
                                        # Activate Elaborate Variable Rejection (EVREJ) or not.
                                        # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE 

        -EVREJ_RATIO_THREH [5.0]        # EVREJ kicks out the variables by the photometric uncertainties
                                        # (SExtractor FLUX_AUTO) measured on reference and science.
                                        # The flux change of a stationary object from reference to science 
                                        # is a predictable probability ditribution. EVREJ will reject the 
                                        # -EVREJ_RATIO_THREH * SIGMA outliers of the probability distribution.
                                        # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE 
        
        -EVREJ_SAFE_MAGDEV [0.04]       # A protection to avoid EVREJ overkilling. EVREJ would not reject
                                        # a source if the difference of instrument magnitudes is deviated
                                        # from the median level within -EVREJ_SAFE_MAGDEV magnitude.
                                        # NOTE: ONLY WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE 
                                        #
                                        # @ Why does EVREJ overkill?
                                        # i) The calculated flux change is entangled with the error of 
                                        #    the constant photometric scaling used by EVREJ.
                                        # ii) SExtractor FLUXERR can be in accurate, especially at the bright end.
                                        # see more details in sfft.Auto_SparsePrep.HoughAutoMask

        -StarExt_iter [4]               # Our image mask is determined by the SExtractor check image SEGMENTATION 
                                        # of the selected sources. note that some pixels (e.g., outskirt region of a galaxy) 
                                        # harbouring signal may be not SExtractor-detectable due to the nature of 
                                        # the thresholding-based detection method (see NoiseChisel paper for more details). 
                                        # we want to include the missing light at outskirt region to contribute to
                                        # parameter-solving process, then a simple mask dilation is introduced.
                                        # -StarExt_iter means the iteration times of the dilation process.
                                        # NOTE: IT WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE :::: mask dilation
                                        # NOTE: IT WORKS FOR Auto-Sparse-Prep [SEMI-AUTO] MODE :::: mask dilation

        -XY_PriorBan_Queue [[]]         # A queue of -XY_PriorBan in sfft.Easy_SparsePacket. 
                                        # Recall -XY_PriorBan: 
                                        # a Numpy array of pixels coordinates, with shape (N, 2) (e.g., [[x0, y0], [x1, y1], ...])
                                        # this allows us to feed the prior knowledge about the varibility cross the field.
                                        # if you already get a list of variables (transients) and would like SFFT not to select
                                        # them to solve the subtraction, you can tell SFFT their coordinates through -XY_PriorBan.
                                        # NOTE: IT WORKS FOR Auto-Sparse-Prep [HOUGH-AUTO] MODE :::: mask refinement
                                        # NOTE: IT WORKS FOR Auto-Sparse-Prep [SEMI-AUTO] MODE :::: mask refinement     

        # ----------------------------- SFFT Subtraction --------------------------------- #

        -ForceConv_Queue [[]]               # A queue of -ForceConv in sfft.Easy_SparsePacket.
                                            # Recall -ForceConv:
                                            # it determines which image will be convolved, can be 'REF', 'SCI' and 'AUTO'.
                                            # -ForceConv = 'AUTO' means SFFT will determine the convolution direction according to 
                                            # FWHM_SCI and FWHM_REF: the image with better seeing will be convolved to avoid deconvolution.

        -GKerHW_Queue [[]]                  # A queue of -GKerHW in sfft.Easy_SparsePacket.
                                            # Recall -GKerHW:
                                            # The given kernel half-width, None means the kernel size will be 
                                            # automatically determined by -KerHWRatio (to be seeing-related). 

        -KerHWRatio [2.0]                   # The ratio between FWHM and the kernel half-width
                                            # KerHW = int(KerHWRatio * Max(FWHM_REF, FWHM_SCI))

        -KerHWLimit [(2, 20)]               # The lower & upper bounds for kernel half-width 
                                            # KerHW is updated as np.clip(KerHW, KerHWLimit[0], KerHWLimit[1]) 
                                            # Remarks: this is useful for a survey since it can constrain the peak GPU memory usage.

        -KerPolyOrder [2]                   # Polynomial degree of kernel spatial variation, can be [0,1,2,3].

        -BGPolyOrder [0]                    # Polynomial degree of differential-background spatial variation, can be [0,1,2,3].
                                            # NOTE: this argument is trivial for Sparse-Flavor SFFT as input images have been sky subtracted.
                                            #       simply use a flat differential background (it will very close to 0 in the sfft solution). 

        -ConstPhotRatio [True]              # Constant photometric ratio between images? can be True or False
                                            # ConstPhotRatio = True: the sum of convolution kernel is restricted 
                                            #   to be a constant across the field. 
                                            # ConstPhotRatio = False: the flux scaling between images is modeled
                                            #   by a polynomial with degree -KerPolyOrder.

        -MaskSatContam [False]              # Mask saturation-contaminated regions on difference image ? can be True or False
                                            # NOTE the pixels enclosed in the regions are replaced by NaN.
        
        # ----------------------------- Post Subtraction --------------------------------- #

        -PostAnomalyCheck [False]           # WARNING: PAC REQUIRES CORRECT GAIN IN FITS HEADER!
                                            # Post Anomaly Check (PAC) is to find the missing variables in the source selection 
                                            # on the difference image. Like EVREJ, it depends on the photometric uncertainties 
                                            # given by SExtractor FLUXERR_AUTO of reference and science image.
                                            # As image subtraction can get a more accurate photometric scaling between reference 
                                            # and science than that in EVREJ, PA is a even more 'elaborate' way to identify variables.
                                            # NOTE: one can make use of the Post-Anomalies as the Prior-Ban-Sources 
                                            #       to refine the subtraction (run second-time subtraction).

        -PAC_RATIO_THRESH [3.0]             # PA will identify the -PAC_RATIO_THRESH * SIGMA outliers of the 
                                            # probability distribution as missing variables.

        # ----------------------------- Input & Output --------------------------------- #

        -FITS_REF_Queue [[]]                # A queue of -FITS_REF in sfft.Easy_SparsePacket.
                                            # Recall -FITS_REF: File path of input reference image.

        -FITS_SCI_Queue [[]]                # A queue of -FITS_SCI in sfft.Easy_SparsePacket.
                                            # Recall -FITS_SCI: File path of input science image.

        -FITS_DIFF_Queue [[]]               # A queue of -FITS_DIFF in sfft.Easy_SparsePacket.
                                            # Recall -FITS_DIFF:File path of output difference image.

        -FITS_Solution_Queue [[]]           # A queue of -FITS_Solution in sfft.Easy_SparsePacket.
                                            # Recall -FITS_Solution: File path of the solution of the linear system. 
                                            # it is an array of (..., a_ijab, ... b_pq, ...).

        # ----------------------------- Miscellaneous --------------------------------- #
        
        -CLEAN_GPU_MEMORY [False]           # Clean GPU memory or not after all subtractions done in a GPU device.

        -VERBOSE_LEVEL [2]                  # The level of verbosity, can be [0, 1, 2]
                                            # 0/1/2: QUIET/NORMAL/FULL mode

        # Important Notice:
        #
        # a): if reference is convolved in SFFT, then DIFF = SCI - Convolved_REF.
        #     [difference image is expected to have PSF & flux zero-point consistent with science image]
        #     e.g., -ForceConv='REF' or -ForceConv='AUTO' when reference has better seeing.
        #
        # b): if science is convolved in SFFT, then DIFF = Convolved_SCI - REF
        #     [difference image is expected to have PSF & flux zero-point consistent with reference image]
        #     e.g., -ForceConv='SCI' or -ForceConv='AUTO' when science has better seeing.
        #
        # Remarks: this convention is to guarantee that transients emerge on science image always 
        #          show a positive signal on difference images.

        """
        
        self.FITS_REF_Queue = FITS_REF_Queue
        self.FITS_SCI_Queue = FITS_SCI_Queue
        self.NUM_TASK = len(FITS_SCI_Queue)

        self.KerHWRatio = KerHWRatio
        self.KerHWLimit = KerHWLimit
        self.KerPolyOrder = KerPolyOrder
        self.BGPolyOrder = BGPolyOrder
        self.ConstPhotRatio = ConstPhotRatio
        self.MaskSatContam = MaskSatContam
        
        self.GAIN_KEY = GAIN_KEY
        self.SATUR_KEY = SATUR_KEY
        self.BACK_TYPE = BACK_TYPE
        self.BACK_VALUE = BACK_VALUE
        self.BACK_SIZE = BACK_SIZE
        self.BACK_FILTERSIZE = BACK_FILTERSIZE
        
        self.DETECT_THRESH = DETECT_THRESH
        self.ANALYSIS_THRESH = ANALYSIS_THRESH
        self.DETECT_MINAREA = DETECT_MINAREA
        self.DETECT_MAXAREA = DETECT_MAXAREA
        self.DEBLEND_MINCONT = DEBLEND_MINCONT
        self.BACKPHOTO_TYPE = BACKPHOTO_TYPE
        self.ONLY_FLAGS = ONLY_FLAGS
        self.BoundarySIZE = BoundarySIZE

        self.Hough_MINFR = Hough_MINFR
        self.Hough_PeakClip = Hough_PeakClip
        self.BeltHW = BeltHW
        self.PointSource_MINELLIP = PointSource_MINELLIP

        self.MatchTol = MatchTol
        self.MatchTolFactor = MatchTolFactor

        self.COARSE_VAR_REJECTION = COARSE_VAR_REJECTION
        self.CVREJ_MAGD_THRESH = CVREJ_MAGD_THRESH
        self.ELABO_VAR_REJECTION = ELABO_VAR_REJECTION
        self.EVREJ_RATIO_THREH = EVREJ_RATIO_THREH
        self.EVREJ_SAFE_MAGDEV = EVREJ_SAFE_MAGDEV
        
        self.StarExt_iter = StarExt_iter

        self.PostAnomalyCheck = PostAnomalyCheck
        self.PAC_RATIO_THRESH = PAC_RATIO_THRESH

        self.CLEAN_GPU_MEMORY = CLEAN_GPU_MEMORY
        self.VERBOSE_LEVEL = VERBOSE_LEVEL

        if FITS_DIFF_Queue == []:
            if self.VERBOSE_LEVEL in [2]:
                _warn_message = 'Argument FITS_DIFF_Queue is EMPTY then Use default [...None...] instead!'
                warnings.warn('\nMeLOn WARNING: %s' %_warn_message)
            self.FITS_DIFF_Queue = [None] * self.NUM_TASK
        else: self.FITS_DIFF_Queue = FITS_DIFF_Queue
        
        if FITS_Solution_Queue == []:
            if self.VERBOSE_LEVEL in [2]:
                _warn_message = 'Argument FITS_Solution_Queue is EMPTY then Use default [...None...] instead!'
                warnings.warn('\nMeLOn WARNING: %s' %_warn_message)
            self.FITS_Solution_Queue = [None] * self.NUM_TASK
        else: self.FITS_Solution_Queue = FITS_Solution_Queue
        
        if ForceConv_Queue == []:
            if self.VERBOSE_LEVEL in [0, 1, 2]:
                _warn_message = 'Argument ForceConv_Queue is EMPTY then Use default [...AUTO...] instead!'
                warnings.warn('\nMeLOn WARNING: %s' %_warn_message)
            self.ForceConv_Queue = ['AUTO'] * self.NUM_TASK
        else: self.ForceConv_Queue = ForceConv_Queue

        if GKerHW_Queue == []:
            if self.VERBOSE_LEVEL in [2]:
                _warn_message = 'Argument GKerHW_Queue is EMPTY then Use default [...None...] instead!'
                warnings.warn('\nMeLOn WARNING: %s' %_warn_message)
            self.GKerHW_Queue = [None] * self.NUM_TASK
        else: self.GKerHW_Queue = GKerHW_Queue

        if XY_PriorSelect_Queue == []:
            if self.VERBOSE_LEVEL in [2]:
                _warn_message = 'Argument XY_PriorSelect_Queue is EMPTY then Use default [...None...] instead!'
                warnings.warn('\nMeLOn WARNING: %s' %_warn_message)
            self.XY_PriorSelect_Queue = [None] * self.NUM_TASK
        else: self.XY_PriorSelect_Queue = XY_PriorSelect_Queue

        if XY_PriorBan_Queue == []:
            if self.VERBOSE_LEVEL in [2]:
                _warn_message = 'Argument XY_PriorBan_Queue is EMPTY then Use default [...None...] instead!'
                warnings.warn('\nMeLOn WARNING: %s' %_warn_message)
            self.XY_PriorBan_Queue = [None] * self.NUM_TASK
        else: self.XY_PriorBan_Queue = XY_PriorBan_Queue

        if self.PostAnomalyCheck:
            if self.VERBOSE_LEVEL in [2]:    
                warnings.warn('\nMeLOn REMINDER: Post-Anomaly Check requires correct GAIN values!')

        if self.VERBOSE_LEVEL in [2]:
            warnings.warn('\nMeLOn REMINDER: Input images for sparse-flavor sfft should be SKY-SUBTRACTED!')

    def MESP_Cupy(self, NUM_THREADS_4PREPROC=8, NUM_THREADS_4SUBTRACT=1, CUDA_DEVICES_4SUBTRACT=['0'], \
        TIMEOUT_4PREPROC_EACHTASK=300, TIMEOUT_4SUBTRACT_EACHTASK=300):

        import cupy as cp
        
        # * define task-oriented dictionaries of status-bar and products
        #   status 0: Initial State
        #   status 1: Preprocessing SUCCESS || status -1: Preprocessing FAIL
        #   status 2: Subtraction SUCCESS || status -2: Subtraction FAIL
        #   status 32: Preprocessing is IN-PROGRESS ...
        #   status 64: Subtraction is IN-PROGRESS ...
        #   NOTE: Subtraction would not be triggered when Preprocessing fails
    
        _RATIO = self.NUM_TASK / NUM_THREADS_4PREPROC
        if _RATIO > 4.0:
            if self.VERBOSE_LEVEL in [0, 1, 2]:
                _warn_message = 'THIS FUNCTION IS NOT OPTIMIZED FOR RATIO OF '
                _warn_message += 'NUM_TASK / NUM_THREADS_4PREPROC BEING TO HIGH [%.1f > 4.0]!' %_RATIO
                warnings.warn('\nMeLOn WARNING: %s' %_warn_message)

        BACKEND_4SUBTRACT = 'Cupy'
        taskidx_lst = np.arange(self.NUM_TASK)
        assert NUM_THREADS_4SUBTRACT == len(CUDA_DEVICES_4SUBTRACT)

        DICT_STATUS_BAR = {'TASK-[%d]' %taskidx: 0 for taskidx in taskidx_lst}
        DICT_PRODUCTS = {'TASK-[%d]' %taskidx: {} for taskidx in taskidx_lst}
        lock = threading.RLock()

        # * define the function for preprocessing (see sfft.EasySparsePacket)
        def FUNC_4PREPROC(INDEX_THREAD_4PREPROC):
            
            THREAD_ALIVE_4PREPROC = True
            while THREAD_ALIVE_4PREPROC:
                with lock:
                    STATUS_SNAPSHOT = np.array([DICT_STATUS_BAR['TASK-[%d]' %taskidx] for taskidx in taskidx_lst])
                    OPEN_MASK_4PREPROC = STATUS_SNAPSHOT == 0
                    if np.sum(OPEN_MASK_4PREPROC) > 0:
                        taskidx_acquired = taskidx_lst[OPEN_MASK_4PREPROC][0]
                        DICT_STATUS_BAR['TASK-[%d]' %taskidx_acquired] = 32

                        if self.VERBOSE_LEVEL in [2]:
                            _message = 'THREAD-4PREPROC-[%d] ACQUIRED TASK-[%d] ' %(INDEX_THREAD_4PREPROC, taskidx_acquired)
                            _message += '(Sparse-Flavor Auto Preprocessing)!'
                            print('\nMeLOn CheckPoint: %s' %_message)
                    else: 
                        THREAD_ALIVE_4PREPROC = False
                        taskidx_acquired = None

                        if self.VERBOSE_LEVEL in [2]:
                            _message = 'THREAD-4PREPROC-[%d] TERMINATED ' %INDEX_THREAD_4PREPROC
                            _message += 'SINCE NO MORE TASK WILL BE ALLOCATED '
                            _message += '(Sparse-Flavor Auto Preprocessing)!'
                            print('\nMeLOn CheckPoint: %s' %_message)

                if taskidx_acquired is not None:
                    FITS_REF = self.FITS_REF_Queue[taskidx_acquired]
                    FITS_SCI = self.FITS_SCI_Queue[taskidx_acquired]
                    XY_PriorBan = self.XY_PriorBan_Queue[taskidx_acquired]
                    XY_PriorSelect = self.XY_PriorSelect_Queue[taskidx_acquired]

                    try:
                        with TimeoutAfter(timeout=TIMEOUT_4PREPROC_EACHTASK, exception=TimeoutError):

                            # ** perform Auto-Sparse-Prep
                            _ASP = Auto_SparsePrep(FITS_REF=FITS_REF, FITS_SCI=FITS_SCI, GAIN_KEY=self.GAIN_KEY, \
                                SATUR_KEY=self.SATUR_KEY, BACK_TYPE=self.BACK_TYPE, BACK_VALUE=self.BACK_VALUE, \
                                BACK_SIZE=self.BACK_SIZE, BACK_FILTERSIZE=self.BACK_FILTERSIZE, \
                                DETECT_THRESH=self.DETECT_THRESH, ANALYSIS_THRESH=self.ANALYSIS_THRESH, \
                                DETECT_MINAREA=self.DETECT_MINAREA, DETECT_MAXAREA=self.DETECT_MAXAREA, \
                                DEBLEND_MINCONT=self.DEBLEND_MINCONT, BACKPHOTO_TYPE=self.BACKPHOTO_TYPE, \
                                ONLY_FLAGS=self.ONLY_FLAGS, BoundarySIZE=self.BoundarySIZE, \
                                VERBOSE_LEVEL=self.VERBOSE_LEVEL)

                            if XY_PriorSelect is None:
                                IMAGE_MASK_METHOD = 'HOUGH-AUTO'
                                if self.VERBOSE_LEVEL in [1, 2]:
                                    _message = 'THREAD-4PREPROC-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4PREPROC, taskidx_acquired)
                                    _message += 'TRIGGER Sparse-Flavor Auto Preprocessing [%s] MODE!' %IMAGE_MASK_METHOD
                                    print('MeLOn CheckPoint: %s' %_message)
                                
                                SFFTPrepDict = _ASP.HoughAutoMask(Hough_MINFR=self.Hough_MINFR, \
                                    Hough_PeakClip=self.Hough_PeakClip, BeltHW=self.BeltHW, \
                                    PointSource_MINELLIP=self.PointSource_MINELLIP, MatchTol=self.MatchTol, \
                                    MatchTolFactor=self.MatchTolFactor, ELABO_VAR_REJECTION=self.ELABO_VAR_REJECTION, \
                                    EVREJ_RATIO_THREH=self.EVREJ_RATIO_THREH, EVREJ_SAFE_MAGDEV=self.EVREJ_SAFE_MAGDEV, \
                                    StarExt_iter=self.StarExt_iter, XY_PriorBan=XY_PriorBan)
                            else:
                                IMAGE_MASK_METHOD = 'SEMI-AUTO'
                                if self.VERBOSE_LEVEL in [1, 2]:
                                    _message = 'THREAD-4PREPROC-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4PREPROC, taskidx_acquired)
                                    _message += 'TRIGGER Sparse-Flavor Auto Preprocessing [%s] MODE!' %IMAGE_MASK_METHOD
                                    print('MeLOn CheckPoint: %s' %_message)

                                SFFTPrepDict = _ASP.SemiAutoMask(XY_PriorSelect=XY_PriorSelect, MatchTol=self.MatchTol, \
                                    MatchTolFactor=self.MatchTolFactor, StarExt_iter=self.StarExt_iter)

                        NEW_STATUS = 1
                        if self.VERBOSE_LEVEL in [1, 2]:
                            _message = 'THREAD-4PREPROC-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4PREPROC, taskidx_acquired)
                            _message += 'SUCCESSFUL Sparse-Flavor Auto Preprocessing!'
                            print('\nMeLOn CheckPoint: %s' %_message)

                    except Exception as e:
                        NEW_STATUS = -1
                        SFFTPrepDict = None
                        if self.VERBOSE_LEVEL in [0, 1, 2]:
                            print(e, file=sys.stderr)
                            _error_message = 'THREAD-4PREPROC-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4PREPROC, taskidx_acquired)
                            _error_message += 'FAILED Sparse-Flavor Auto Preprocessing!'
                            print('\nMeLOn ERROR: %s' %_error_message)

                    with lock:
                        # NOTE: IN FACT, THERE IS NO RISK HERE EVEN WITHOUT LOCK.
                        DICT_STATUS_BAR['TASK-[%d]' %taskidx_acquired] = NEW_STATUS
                        DICT_PRODUCTS['TASK-[%d]' %taskidx_acquired]['SFFTPrepDict'] = SFFTPrepDict

            return None

        # * define the function for sfft subtraction (see sfft.EasySparsePacket)
        def FUNC_4SUBTRCT(INDEX_THREAD_4SUBTRACT):
            
            CUDA_DEVICE_4SUBTRACT = CUDA_DEVICES_4SUBTRACT[INDEX_THREAD_4SUBTRACT]
            device = cp.cuda.Device(int(CUDA_DEVICE_4SUBTRACT))
            device.use()

            THREAD_ALIVE_4SUBTRCT = True
            GPU_MEMORY_CLEANED = True

            while THREAD_ALIVE_4SUBTRCT:
                SHORT_PAUSE = False

                with lock:
                    STATUS_SNAPSHOT = np.array([DICT_STATUS_BAR['TASK-[%d]' %taskidx] for taskidx in taskidx_lst])
                    OPEN_MASK_4SUBTRACT = STATUS_SNAPSHOT == 1
                    WAIT_MASK_4SUBTRACT = np.in1d(STATUS_SNAPSHOT, [0, 32])

                    if np.sum(OPEN_MASK_4SUBTRACT) > 0:
                        taskidx_acquired = taskidx_lst[OPEN_MASK_4SUBTRACT][0]
                        DICT_STATUS_BAR['TASK-[%d]' %taskidx_acquired] = 64

                        if self.VERBOSE_LEVEL in [2]:
                            _message = 'THREAD-4SUBTRACT-[%d] ACQUIRED TASK-[%d] ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                            _message += '(SFFT-SUBTRACTION)!'
                            print('\nMeLOn CheckPoint: %s' %_message)

                    elif np.sum(WAIT_MASK_4SUBTRACT) > 0:
                        SHORT_PAUSE = True
                        taskidx_acquired = None

                    else:
                        THREAD_ALIVE_4SUBTRCT = False
                        taskidx_acquired = None

                        if self.VERBOSE_LEVEL in [2]:
                            _message = 'THREAD-4SUBTRACT-[%d] TERMINATED ' %INDEX_THREAD_4SUBTRACT
                            _message += 'SINCE NO MORE TASK WILL BE ALLOCATED '
                            _message += '(SFFT-SUBTRACTION)!'
                            print('\nMeLOn CheckPoint: %s' %_message)

                if SHORT_PAUSE:
                    # run a short sleep, otherwise this thread may always hold the lock by the while loop
                    time.sleep(0.01)
                
                if taskidx_acquired is not None:

                    FITS_REF = self.FITS_REF_Queue[taskidx_acquired]
                    FITS_SCI = self.FITS_SCI_Queue[taskidx_acquired]
                    FITS_DIFF = self.FITS_DIFF_Queue[taskidx_acquired]
                    FITS_Solution = self.FITS_Solution_Queue[taskidx_acquired]
                    ForceConv = self.ForceConv_Queue[taskidx_acquired]
                    GKerHW = self.GKerHW_Queue[taskidx_acquired]

                    with lock:
                        # NOTE: IN FACT, THERE IS NO RISK HERE EVEN WITHOUT LOCK.
                        SFFTPrepDict = DICT_PRODUCTS['TASK-[%d]' %taskidx_acquired]['SFFTPrepDict']
                    
                    try:
                        with TimeoutAfter(timeout=TIMEOUT_4SUBTRACT_EACHTASK, exception=TimeoutError):                    
                            
                            # * Determine ConvdSide & KerHW
                            FWHM_REF = SFFTPrepDict['FWHM_REF']
                            FWHM_SCI = SFFTPrepDict['FWHM_SCI']

                            assert ForceConv in ['AUTO', 'REF', 'SCI']
                            if ForceConv == 'AUTO':
                                if FWHM_SCI >= FWHM_REF: ConvdSide = 'REF'
                                else: ConvdSide = 'SCI'
                            else: ConvdSide = ForceConv

                            if GKerHW is None:
                                FWHM_La = np.max([FWHM_REF, FWHM_SCI])
                                KerHW = int(np.clip(self.KerHWRatio * FWHM_La, \
                                                    self.KerHWLimit[0], self.KerHWLimit[1]))
                            else: KerHW = GKerHW
                            
                            # * Compile Functions in SFFT Subtraction
                            

                            PixA_REF = SFFTPrepDict['PixA_REF']
                            PixA_SCI = SFFTPrepDict['PixA_SCI']
                            
                            if self.VERBOSE_LEVEL in [1, 2]:
                                _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                _message += 'TRIGGER Function Compilations of SFFT-SUBTRACTION'
                                print('MeLOn CheckPoint: %s' %_message)

                            Tcomp_start = time.time()
                            GPU_MEMORY_CLEANED = False
                            SFFTConfig = SingleSFFTConfigure.SSC(NX=PixA_REF.shape[0], NY=PixA_REF.shape[1], \
                                KerHW=KerHW, KerPolyOrder=self.KerPolyOrder, BGPolyOrder=self.BGPolyOrder, \
                                ConstPhotRatio=self.ConstPhotRatio, BACKEND_4SUBTRACT=BACKEND_4SUBTRACT, \
                                VERBOSE_LEVEL=self.VERBOSE_LEVEL)

                            if self.VERBOSE_LEVEL in [1, 2]:
                                _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                _message += 'Function Compilations of SFFT-SUBTRACTION '
                                _message += 'TAKES [%.3f s]!' %(time.time() - Tcomp_start)
                                print('\nMeLOn Report: %s' %_message)

                            # * Perform SFFT Subtraction
                            

                            SatMask_REF = SFFTPrepDict['REF-SAT-Mask']
                            SatMask_SCI = SFFTPrepDict['SCI-SAT-Mask']
                            NaNmask_U = SFFTPrepDict['Union-NaN-Mask']
                            PixA_mREF = SFFTPrepDict['PixA_mREF']
                            PixA_mSCI = SFFTPrepDict['PixA_mSCI']

                            if ConvdSide == 'REF':
                                PixA_mI, PixA_mJ = PixA_mREF, PixA_mSCI
                                if NaNmask_U is not None:
                                    PixA_I, PixA_J = PixA_REF.copy(), PixA_SCI.copy()
                                    PixA_I[NaNmask_U] = PixA_mI[NaNmask_U]
                                    PixA_J[NaNmask_U] = PixA_mJ[NaNmask_U]
                                else: PixA_I, PixA_J = PixA_REF, PixA_SCI
                                if self.MaskSatContam: 
                                    ContamMask_I = SatMask_REF
                                    ContamMask_J = SatMask_SCI
                                else: ContamMask_I = None

                            if ConvdSide == 'SCI':
                                PixA_mI, PixA_mJ = PixA_mSCI, PixA_mREF
                                if NaNmask_U is not None:
                                    PixA_I, PixA_J = PixA_SCI.copy(), PixA_REF.copy()
                                    PixA_I[NaNmask_U] = PixA_mI[NaNmask_U]
                                    PixA_J[NaNmask_U] = PixA_mJ[NaNmask_U]
                                else: PixA_I, PixA_J = PixA_SCI, PixA_REF
                                if self.MaskSatContam: 
                                    ContamMask_I = SatMask_SCI
                                    ContamMask_J = SatMask_REF
                                else: ContamMask_I = None

                            if self.VERBOSE_LEVEL in [1, 2]:
                                _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                _message += 'TRIGGER Function Executions of SFFT-SUBTRACTION!'
                                print('MeLOn CheckPoint: %s' %_message)

                            Tsub_start = time.time()
                            _tmp = GeneralSFFTSubtract.GSS(PixA_I=PixA_I, PixA_J=PixA_J, PixA_mI=PixA_mI, PixA_mJ=PixA_mJ, \
                                SFFTConfig=SFFTConfig, ContamMask_I=ContamMask_I, BACKEND_4SUBTRACT=BACKEND_4SUBTRACT, \
                                VERBOSE_LEVEL=self.VERBOSE_LEVEL)
                            
                            Solution, PixA_DIFF, ContamMask_CI = _tmp
                            if self.VERBOSE_LEVEL in [1, 2]:
                                _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                _message += 'Function Executions of SFFT-SUBTRACTION '
                                _message += 'TAKES [%.3f s]!' %(time.time() - Tsub_start)
                                print('\nMeLOn Report: %s' %_message)

                            # ** adjust the sign of difference image 
                            #    NOTE: Our convention is to make the transients on SCI always show themselves as positive signal on DIFF
                            #    (a) when REF is convolved, DIFF = SCI - Conv(REF)
                            #    (b) when SCI is convolved, DIFF = Conv(SCI) - REF

                            if ConvdSide == 'SCI':
                                PixA_DIFF = -PixA_DIFF

                            if self.VERBOSE_LEVEL in [1, 2]:
                                if ConvdSide == 'REF':
                                    _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                    _message += 'Reference Image is Convolved in SFFT-SUBTRACTION [DIFF = SCI - Conv(REF)]!'
                                    print('MeLOn CheckPoint: %s' %_message)

                                if ConvdSide == 'SCI':
                                    _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                    _message += 'Science Image is Convolved in SFFT-SUBTRACTION [DIFF = Conv(SCI) - REF]!'
                                    print('MeLOn CheckPoint: %s' %_message)

                            # ** estimate the flux scaling through the convolution of image subtraction
                            #    NOTE: if photometric ratio is not constant (ConstPhotRatio=False), we measure of a grid
                            #          of coordinates to estimate the flux scaling and its fluctuation (polynomial form).
                            #    NOTE: here we set the tile-size of the grid about 64 x 64 pix, but meanwhile, 
                            #          we also require the tiles should no less than 6 along each axis.

                            N0, N1 = SFFTConfig[0]['N0'], SFFTConfig[0]['N1']
                            L0, L1 = SFFTConfig[0]['L0'], SFFTConfig[0]['L1']
                            DK, Fpq = SFFTConfig[0]['DK'], SFFTConfig[0]['Fpq']

                            if self.ConstPhotRatio:
                                SFFT_FSCAL_NSAMP = 1
                                XY_q = np.array([[N0/2.0, N1/2.0]]) + 0.5
                                RFS = Realize_FluxScaling(XY_q=XY_q)
                                SFFT_FSCAL_q = RFS.FromArray(Solution=Solution, N0=N0, N1=N1, L0=L0, L1=L1, DK=DK, Fpq=Fpq)
                                SFFT_FSCAL_MEAN, SFFT_FSCAL_SIG = SFFT_FSCAL_q[0], 0.0
                            
                            if not self.ConstPhotRatio:
                                TILESIZE_X, TILESIZE_Y = 64, 64    # FIXME emperical/arbitrary values
                                NTILE_X = np.max([round(N0/TILESIZE_X), 6])
                                NTILE_Y = np.max([round(N1/TILESIZE_Y), 6])
                                GX = np.linspace(0.5, N0+0.5, NTILE_X+1)
                                GY = np.linspace(0.5, N1+0.5, NTILE_Y+1)
                                YY, XX = np.meshgrid(GY, GX)
                                XY_q = np.array([XX.flatten(), YY.flatten()]).T
                                SFFT_FSCAL_NSAMP = XY_q.shape[0]
                                
                                RFS = Realize_FluxScaling(XY_q=XY_q)
                                SFFT_FSCAL_q = RFS.FromArray(Solution=Solution, N0=N0, N1=N1, L0=L0, L1=L1, DK=DK, Fpq=Fpq)
                                SFFT_FSCAL_MEAN, SFFT_FSCAL_SIG = np.mean(SFFT_FSCAL_q), np.std(SFFT_FSCAL_q)

                            PHOT_FSCAL = 10**(SFFTPrepDict['MAG_OFFSET']/-2.5)  # FLUX_SCI/FLUX_REF
                            if ConvdSide == 'SCI': PHOT_FSCAL = 1.0/PHOT_FSCAL

                            if self.VERBOSE_LEVEL in [1, 2]:
                                _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                _message += 'The Flux Scaling through the Convolution of SFFT-SUBTRACTION '
                                _message += '[%.6f +/- %.6f] from [%d] positions!\n' %(SFFT_FSCAL_MEAN, SFFT_FSCAL_SIG, SFFT_FSCAL_NSAMP)
                                _message += 'P.S. The approximated Flux Scaling from Photometry [%.6f].' %PHOT_FSCAL
                                print('MeLOn CheckPoint: %s' %_message)
                            
                            # ** run post anomaly check (PAC)
                            if self.PostAnomalyCheck:

                                """
                                # Remarks on the Post Anomaly Check (PAC)
                                # [1] One may notice that Elaborate Variable Rejection (EVREJ) also use SExtractor 
                                #     photometric errors to identify variables. However, PAC should be more accurate than EVREJ.
                                #     NOTE: Bright transients (compared with its host) can be also identified by PAC.
                                #
                                # [2] distribution mean: image subtraction is expected to have better accuracy on the determination 
                                #     of the photometric scaling, compared with the constant MAG_OFFSET in EVREJ. As a result,
                                #     the flux change counted on the difference image is likely more 'real' than that in EVREJ.
                                #
                                # [3] distribution variance: SExtractor FLUXERR_AUTO can be inaccurate, and the convolution effect
                                #     on the photometric errors has been simplified as multiplying an average flux scaling on an image.
                                #
                                """

                                if self.VERBOSE_LEVEL in [2]:
                                    warnings.warn('\nMeLOn REMINDER: Post-Anomaly Check requires CORRECT GAIN in FITS HEADER!')

                                AstSEx_SS = SFFTPrepDict['SExCatalog-SubSource']
                                SFFTLmap = SFFTPrepDict['SFFT-LabelMap']
                                
                                # ** Only consider the Non-Prior-Banned SubSources
                                if 'MASK_PriorBan' in AstSEx_SS.colnames:
                                    nPBMASK_SS = ~np.array(AstSEx_SS['MASK_PriorBan'])
                                    AstSEx_vSS = AstSEx_SS[nPBMASK_SS]
                                else: AstSEx_vSS = AstSEx_SS
                                
                                # ** Estimate expected variance of Non-Prior-Banned SubSources on the difference
                                FLUXERR_vSSr = np.array(AstSEx_vSS['FLUXERR_AUTO_REF'])
                                FLUXERR_vSSs = np.array(AstSEx_vSS['FLUXERR_AUTO_SCI'])

                                if ConvdSide == 'REF':
                                    sFLUXERR_vSSr = FLUXERR_vSSr * SFFT_FSCAL_MEAN
                                    ExpDVAR_vSS = sFLUXERR_vSSr**2 + FLUXERR_vSSs**2

                                if ConvdSide == 'SCI':
                                    sFLUXERR_vSSs = FLUXERR_vSSs * SFFT_FSCAL_MEAN
                                    ExpDVAR_vSS = FLUXERR_vSSr**2 + sFLUXERR_vSSs**2

                                # ** Measure the ratios of Non-Prior-Banned SubSources on the difference for PAC
                                SEGL_vSS = np.array(AstSEx_vSS['SEGLABEL']).astype(int)
                                DFSUM_vSS = ndimage.labeled_comprehension(PixA_DIFF, SFFTLmap, SEGL_vSS, np.sum, float, 0.0)
                                RATIO_vSS = DFSUM_vSS / np.clip(np.sqrt(ExpDVAR_vSS), a_min=1e-8, a_max=None)
                                PAMASK_vSS = np.abs(RATIO_vSS) > self.PAC_RATIO_THRESH

                                if self.VERBOSE_LEVEL in [1, 2]:
                                    _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                    _message += 'Identified [%d] PostAnomaly SubSources [> %.2f sigma] ' %(np.sum(PAMASK_vSS), self.PAC_RATIO_THRESH)
                                    _message += 'out of [%d] Non-Prior-Banned SubSources!\n' %(len(AstSEx_vSS))
                                    _message += 'P.S. there are [%d] Prior-Banned SubSources!' %(len(AstSEx_SS) - len(AstSEx_vSS))
                                    print('\nMeLOn CheckPoint: %s' %_message)
                                
                                # ** Record the results (decorate AstSEx_SS in SFFTPrepDict)
                                if 'MASK_PriorBan' in AstSEx_SS.colnames:
                                    ExpDVAR_SS = np.nan * np.ones(len(AstSEx_SS))       # NOTE Prior-Ban is trivial NaN
                                    DFSUM_SS = np.nan * np.ones(len(AstSEx_SS))         # NOTE Prior-Ban is trivial NaN
                                    RATIO_SS = np.nan * np.ones(len(AstSEx_SS))         # NOTE Prior-Ban is trivial NaN
                                    PAMASK_SS = np.zeros(len(AstSEx_SS)).astype(bool)   # NOTE Prior-Ban is trivial False
                                    
                                    ExpDVAR_SS[nPBMASK_SS] = ExpDVAR_vSS
                                    DFSUM_SS[nPBMASK_SS] = DFSUM_vSS
                                    RATIO_SS[nPBMASK_SS] = RATIO_vSS
                                    PAMASK_SS[nPBMASK_SS] = PAMASK_vSS
                                else: 
                                    ExpDVAR_SS = ExpDVAR_vSS
                                    DFSUM_SS = DFSUM_vSS
                                    RATIO_SS = RATIO_vSS
                                    PAMASK_SS = PAMASK_vSS

                                AstSEx_SS.add_column(Column(ExpDVAR_SS, name='ExpDVAR_PostAnomaly'))
                                AstSEx_SS.add_column(Column(DFSUM_SS, name='DFSUM_PostAnomaly'))
                                AstSEx_SS.add_column(Column(RATIO_SS, name='RATIO_PostAnomaly'))
                                AstSEx_SS.add_column(Column(PAMASK_SS, name='MASK_PostAnomaly'))
                            
                            # ** final tweaks on the difference image
                            if NaNmask_U is not None:
                                PixA_DIFF[NaNmask_U] = np.nan   # Mask Union-NaN regions
                            
                            if self.MaskSatContam:
                                ContamMask_DIFF = np.logical_or(ContamMask_CI, ContamMask_J)
                                PixA_DIFF[ContamMask_DIFF] = np.nan   # Mask Saturation-Contaminated regions
                            
                            # * Save difference image
                            if FITS_DIFF is not None:

                                """
                                # Remarks on header of difference image
                                # [1] In general, the FITS header of difference image would mostly be inherited from science image.
                                #
                                # [2] when science image is convolved, we turn to set GAIN_DIFF = GAIN_SCI / SFFT_FSCAL_MEAN (unit: e-/ADU)
                                #     so that one can correctly convert ADU to e- for a transient appears on science image.
                                #     WARNING: for a transient appears on reference image, GAIN_DIFF = GAIN_SCI is correct.
                                #              we should keep in mind that GAIN_DIFF is not an absolute instrumental value.
                                #     WARNING: one can only estimate the Poission noise from the science transient via GIAN_DIFF.
                                #              the Poission noise from the background source (host galaxy) can enhance the 
                                #              background noise at the position of the transient. photometry softwares which 
                                #              only take background noise and transient Possion noise into account would tend 
                                #              to overestimate the SNR of the transient.
                                # 
                                # [3] when science image is convolved, we turn to set SATURA_DIFF = SATURA_SCI * SFFT_FSCAL_MEAN         
                                #     WARNING: it is still not a good idea to use SATURA_DIFF to identify the SATURATION regions
                                #              on difference image. more appropriate way is masking the saturation contaminated 
                                #              regions on difference image (MaskSatContam=True), or alternatively, leave them  
                                #              alone and using AI stamp classifier to reject saturation-related bogus.
                                #
                                """

                                _hdl = fits.open(FITS_SCI)
                                _hdl[0].data[:, :] = PixA_DIFF.T

                                _hdl[0].header['NAME_REF'] = (pa.basename(FITS_REF), 'MeLOn: SFFT')
                                _hdl[0].header['NAME_SCI'] = (pa.basename(FITS_SCI), 'MeLOn: SFFT')
                                _hdl[0].header['FWHM_REF'] = (FWHM_REF, 'MeLOn: SFFT')
                                _hdl[0].header['FWHM_SCI'] = (FWHM_SCI, 'MeLOn: SFFT')
                                _hdl[0].header['KERORDER'] = (self.KerPolyOrder, 'MeLOn: SFFT')
                                _hdl[0].header['BGORDER'] = (self.BGPolyOrder, 'MeLOn: SFFT')
                                _hdl[0].header['CPHOTR'] = (str(self.ConstPhotRatio), 'MeLOn: SFFT')
                                _hdl[0].header['KERHW'] = (KerHW, 'MeLOn: SFFT')
                                _hdl[0].header['CONVD'] = (ConvdSide, 'MeLOn: SFFT')

                                if ConvdSide == 'SCI':
                                    GAIN_SCI = _hdl[0].header[self.GAIN_KEY]
                                    SATUR_SCI = _hdl[0].header[self.SATUR_KEY]
                                    GAIN_DIFF = GAIN_SCI / SFFT_FSCAL_MEAN
                                    SATUR_DIFF = SATUR_SCI * SFFT_FSCAL_MEAN

                                    _hdl[0].header[self.GAIN_KEY] = (GAIN_DIFF, 'MeLOn: SFFT')
                                    _hdl[0].header[self.SATUR_KEY] = (SATUR_DIFF, 'MeLOn: SFFT')

                                _hdl.writeto(FITS_DIFF, overwrite=True)
                                _hdl.close()
                            
                            # * Save solution array
                            if FITS_Solution is not None:
                                phdu = fits.PrimaryHDU()
                                phdu.header['N0'] = (SFFTConfig[0]['N0'], 'MeLOn: SFFT')
                                phdu.header['N1'] = (SFFTConfig[0]['N1'], 'MeLOn: SFFT')
                                phdu.header['DK'] = (SFFTConfig[0]['DK'], 'MeLOn: SFFT')
                                phdu.header['DB'] = (SFFTConfig[0]['DB'], 'MeLOn: SFFT')
                                phdu.header['L0'] = (SFFTConfig[0]['L0'], 'MeLOn: SFFT')
                                phdu.header['L1'] = (SFFTConfig[0]['L1'], 'MeLOn: SFFT')
                                
                                phdu.header['FIJ'] = (SFFTConfig[0]['Fij'], 'MeLOn: SFFT')
                                phdu.header['FAB'] = (SFFTConfig[0]['Fab'], 'MeLOn: SFFT')
                                phdu.header['FPQ'] = (SFFTConfig[0]['Fpq'], 'MeLOn: SFFT')
                                phdu.header['FIJAB'] = (SFFTConfig[0]['Fijab'], 'MeLOn: SFFT')

                                PixA_Solution = Solution.reshape((-1, 1))
                                phdu.data = PixA_Solution.T
                                fits.HDUList([phdu]).writeto(FITS_Solution, overwrite=True)
                        
                        NEW_STATUS = 2
                        if self.VERBOSE_LEVEL in [1, 2]:
                            _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                            _message += 'SUCCESSFUL SFFT-SUBTRACTION!'
                            print('\nMeLOn CheckPoint: %s' %_message)
                    
                    except Exception as e:
                        PixA_DIFF, Solution = None, None
                        SFFT_FSCAL_MEAN, SFFT_FSCAL_SIG = np.nan, np.nan
                        
                        # ** clean GPU memory when the subtraction fails
                        with cp.cuda.Device(int(CUDA_DEVICE_4SUBTRACT)):
                            mempool = cp.get_default_memory_pool()
                            pinned_mempool = cp.get_default_pinned_memory_pool()
                            mempool.free_all_blocks()
                            pinned_mempool.free_all_blocks()
                            
                        GPU_MEMORY_CLEANED = True
                        if self.VERBOSE_LEVEL in [2]:
                            _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                            _message += 'CLEAN GPU-[%s] MEMORY!' %CUDA_DEVICE_4SUBTRACT
                            print('\nMeLOn CheckPoint: %s' %_message)
                        
                        NEW_STATUS = -2
                        if self.VERBOSE_LEVEL in [0, 1, 2]:
                            print(e, file=sys.stderr)
                            _error_message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                            _error_message += 'FAILED SFFT-SUBTRACTION!'
                            print('\nMeLOn ERROR: %s' %_error_message)
                    
                    with lock:
                        DICT_STATUS_BAR['TASK-[%d]' %taskidx_acquired] = NEW_STATUS
                        DICT_PRODUCTS['TASK-[%d]' %taskidx_acquired]['PixA_DIFF'] = PixA_DIFF
                        DICT_PRODUCTS['TASK-[%d]' %taskidx_acquired]['Solution'] = Solution
                        DICT_PRODUCTS['TASK-[%d]' %taskidx_acquired]['SFFT_FSCAL_MEAN'] = SFFT_FSCAL_MEAN
                        DICT_PRODUCTS['TASK-[%d]' %taskidx_acquired]['SFFT_FSCAL_SIG'] = SFFT_FSCAL_SIG
                    
            if self.CLEAN_GPU_MEMORY:
                if not GPU_MEMORY_CLEANED:
                    mempool = cp.get_default_memory_pool()
                    pinned_mempool = cp.get_default_pinned_memory_pool()
                    mempool.free_all_blocks()
                    pinned_mempool.free_all_blocks()
                    
                    if self.VERBOSE_LEVEL in [1, 2]:
                        _message = 'THREAD-4SUBTRACT-[%d]: ' %INDEX_THREAD_4SUBTRACT
                        _message += 'FINAL CLEAN GPU-[%s] MEMORY ' %CUDA_DEVICE_4SUBTRACT
                        print('\nMeLOn CheckPoint: %s' %_message)
                else:
                    if self.VERBOSE_LEVEL in [1, 2]:
                        _message = 'THREAD-4SUBTRACT-[%d]: ' %INDEX_THREAD_4SUBTRACT
                        _message += 'SKIP FINAL CLEAN, SINCE GPU-[%s] MEMORY ALREADY CLEANED' %CUDA_DEVICE_4SUBTRACT
                        print('\nMeLOn CheckPoint: %s' %_message)

            return None
        
        # * trigger the threads for the preprocessing and subtraction
        MyThreadQueue = []
        for INDEX_THREAD_4PREPROC in range(NUM_THREADS_4PREPROC):
            MyThread = threading.Thread(target=FUNC_4PREPROC, args=(INDEX_THREAD_4PREPROC,))
            MyThreadQueue.append(MyThread)
            MyThread.start()

        for INDEX_THREAD_4SUBTRACT in range(NUM_THREADS_4SUBTRACT):
            MyThread = threading.Thread(target=FUNC_4SUBTRCT, args=(INDEX_THREAD_4SUBTRACT,))
            MyThreadQueue.append(MyThread)
            MyThread.start()
  
        for MyThread in MyThreadQueue:
            MyThread.join()
        
        # * show a summary
        if self.VERBOSE_LEVEL in [0, 1, 2]:
            NUM_SUCCESS = np.sum(np.array([DICT_STATUS_BAR[task] for task in DICT_STATUS_BAR]) == 2)
            print('\nMeLOn CheckPoint: SFFT PROCESSED [%d / %d] TASKS SUCCESSFULLY!' %(NUM_SUCCESS, self.NUM_TASK))

        return DICT_STATUS_BAR, DICT_PRODUCTS
